#!/usr/bin/env python3

import sys, pathlib
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

from m1n1.setup import *
from m1n1.hw.dart import DART
from m1n1.utils import *

dart = DART.from_adt(u, "arm-io/dart-dispdfr")

width = 2040
height = 60
stride = 64

def make_fb():
    width = 2040
    fb_size = align_up(width * stride * 4, 8 * 0x4000)
    buf = u.memalign(0x4000, fb_size)
    return (dart.iomap(0, buf, fb_size), buf)


width = 2040
fb_size = align_up(width * stride * 4, 8 * 0x4000)
buf_mask = u.memalign(0x4000, fb_size)
p.memset32(buf_mask, 0, fb_size)
p.memset32(buf_mask, 0xFFFFFFFF, 10*stride)
iova_mask = dart.iomap(0, buf_mask, fb_size)
p.write32(0x228202200, iova_mask)

dart.dump_device(0)

#enable backlight
p.write32(0x228600070,0x8051)
p.write32(0x22860006c,0x229)

#enable fifo and vblank
p.write32(0x228400100, 0x613)

# Color correction magic (idk why but it makes colors actually work)
p.write32(0x228202074, 0x1)
p.write32(0x228202028, 0x10)
p.write32(0x22820202c, 0x1)
p.write32(0x228202020, 0x1)
p.write32(0x228202034, 0x1)


#layer enable
p.write32(0x228204020, 0x1)
p.write32(0x228204068, 0x1)
p.write32(0x2282040b4, 0x1)
p.write32(0x2282040f4, 0x1)
p.write32(0x2282040ac, 0x100000)
p.write32(0x228201038, 0x10001)

#layer size
p.write32(0x228204048, height << 16 | width)
p.write32(0x22820404c, height << 16 | width)
p.write32(0x22820407c, height << 16 | width)
p.write32(0x228204054, height << 16 | width)

#global size
p.write32(0x228201030, height << 16 | width)

#some more color correction
p.write32(0x22820402c, 0x53e4001)

def make_pipe(iova):
    pipe = [
        0x20014038,
        0x2a | (stride * 4),
        0x20014030,
        0x2a | iova,
    ]
    pipe = [0xc0000001 | (len(pipe) << 16)] + pipe
    return pipe

def flush(pipe):
 for i in pipe:
  p.write32(0x2282010c0, i)

def play(f):
    frame = 0
    iova, base = make_fb()
    while True:
        data = f.read(80 * 64 * 4)
        if not data:
            break
        for i in range(25):
            iface.writemem(base + i * 80 * 64 * 4, data)
        flush(make_pipe(iova))
        time.sleep(0.033)
        frame += 1

with open('out.bin', 'rb') as f:
    play(f)



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

from m1n1.setup import *
from m1n1.hw.dart8110 import DART8110, DART8110Regs
from m1n1.hw.prores import *
from m1n1.utils import *
import time


def divroundup(val, div):
    return (val + div - 1) // div


def bswp16(x):
    return (x >> 8) | ((x & 0xFF) << 8)

# ffmpeg -y -i prores-encode-large.png -c:v rawvideo -pix_fmt nv24 prores-encode-large.yuv
im_W = 1920
im_H = 1080
with open('prores-encode-large.yuv', 'rb') as f:
    im_data = f.read()
assert len(im_data) == (im_W*im_H) * 3
image_data_luma = im_data[:im_W*im_H]
image_data_chroma = im_data[im_W*im_H:]

p.pmgr_adt_clocks_enable(f'/arm-io/dart-apr0')
p.pmgr_adt_clocks_enable(f'/arm-io/apr0')

dart = DART8110.from_adt(u, f'/arm-io/dart-apr0')
dart.initialize()

apr_base, _ = u.adt[f'/arm-io/apr0'].get_reg(0)
apr = ProResRegs(u, apr_base)

print(f"Register 0 (ID?) {apr.REG_0x0}")

# TUNABLES
apr.MODE = 0x0
apr.REG_0x118 = apr.REG_0x118.val & ~0x07FF07FF | 0x00000600
apr.REG_0x148 = apr.REG_0x148.val & ~0x00000001 | 0x00000001
apr.REG_0x160 = apr.REG_0x160.val & ~0x800F3FFF | 0x800A04FF
apr.REG_0x164 = apr.REG_0x164.val & ~0x07FF07FF | 0x07800080
apr.REG_0x170 = apr.REG_0x170.val & ~0x800F3FFF | 0x800404FF
apr.REG_0x174 = apr.REG_0x174.val & ~0x07FF07FF | 0x06000080
apr.REG_0x180 = apr.REG_0x180.val & ~0x800F3FFF | 0x800504FF
apr.REG_0x184 = apr.REG_0x184.val & ~0x07FF07FF | 0x06800080
apr.REG_0x190 = apr.REG_0x190.val & ~0x800F3FFF | 0x800004FF
apr.REG_0x194 = apr.REG_0x194.val & ~0x000000FF | 0x00000040
apr.REG_0x1a0 = apr.REG_0x1a0.val & ~0x800F3FFF | 0x800104FF
apr.REG_0x1a4 = apr.REG_0x1a4.val & ~0x000000FF | 0x00000080
apr.REG_0x1b0 = apr.REG_0x1b0.val & ~0x800F3FFF | 0x800204FF
apr.REG_0x1b4 = apr.REG_0x1b4.val & ~0x000000FF | 0x00000040
apr.REG_0x1c0 = apr.REG_0x1c0.val & ~0x800F3FFF | 0x800304FF
apr.REG_0x1c4 = apr.REG_0x1c4.val & ~0x000000FF | 0x00000040
apr.REG_0x1d0 = apr.REG_0x1d0.val & ~0xBC00FF86 | 0xA4000786
apr.REG_0x1d4 = apr.REG_0x1d4.val & ~0x000000FF | 0x00000020
apr.REG_0x1d8 = apr.REG_0x1d8.val & ~0x000000FF | 0x000000FF
apr.REG_0x1dc = apr.REG_0x1dc.val & ~0x00FFFFFF | 0x00928170
apr.REG_0x270 = apr.REG_0x270.val & ~0x800F3FFF | 0x800B08FF
apr.REG_0x274 = apr.REG_0x274.val & ~0x07FF07FF | 0x07000080
apr.REG_0x280 = apr.REG_0x280.val & ~0xFFFFFFC0 | 0x00180000
apr.REG_0x290 = apr.REG_0x290.val & ~0x800F3FFF | 0x800004FF
apr.REG_0x294 = apr.REG_0x294.val & ~0x000000FF | 0x00000080
apr.REG_0x2a0 = apr.REG_0x2a0.val & ~0x800F3FFF | 0x800104FF
apr.REG_0x2a4 = apr.REG_0x2a4.val & ~0x000000FF | 0x00000080
apr.REG_0x2b0 = apr.REG_0x2b0.val & ~0x800F3FFF | 0x800204FF
apr.REG_0x2b4 = apr.REG_0x2b4.val & ~0x000000FF | 0x00000040
apr.REG_0x2c0 = apr.REG_0x2c0.val & ~0x800F3FFF | 0x800304FF
apr.REG_0x2c4 = apr.REG_0x2c4.val & ~0x000000FF | 0x00000040
apr.REG_0x2d0 = apr.REG_0x2d0.val & ~0x802FF04C | 0x80070040
apr.REG_0x2d4 = apr.REG_0x2d4.val & ~0x00000001 | 0x00000000
apr.REG_0x2d8 = apr.REG_0x2d8.val & ~0xFFFF0003 | 0x00FF0003
apr.REG_0x2e0 = apr.REG_0x2e0.val & ~0x07FF07FF | 0x06000040
apr.REG_0x2f8 = apr.REG_0x2f8.val & ~0x802FF04C | 0x80081040
apr.REG_0x2fc = apr.REG_0x2fc.val & ~0x00000001 | 0x00000000
apr.REG_0x300 = apr.REG_0x300.val & ~0xFFFF0003 | 0x00FF0003
apr.REG_0x308 = apr.REG_0x308.val & ~0x07FF07FF | 0x06400040
apr.REG_0x320 = apr.REG_0x320.val & ~0x802FF04C | 0x80092040
apr.REG_0x324 = apr.REG_0x324.val & ~0x00000001 | 0x00000000
apr.REG_0x328 = apr.REG_0x328.val & ~0xFFFF0003 | 0x00FF0003
apr.REG_0x330 = apr.REG_0x330.val & ~0x07FF07FF | 0x06800040
apr.REG_0x350 = apr.REG_0x350.val & ~0x800F3FFF | 0x800B08FF
apr.REG_0x354 = apr.REG_0x354.val & ~0x07FF07FF | 0x076000A0
apr.REG_0x360 = apr.REG_0x360.val & ~0xFFFFFFC0 | 0x00180000
apr.REG_0x370 = apr.REG_0x370.val & ~0x800F3FFF | 0x800604FF
apr.REG_0x374 = apr.REG_0x374.val & ~0x07FF07FF | 0x06C000A0
print("Applied tunables")

# XXX test wrap around behavior
DESC_RING_SZ = 0x4000
desc_ring_phys = u.heap.memalign(0x4000, DESC_RING_SZ)
desc_ring_iova = dart.iomap(0, desc_ring_phys, DESC_RING_SZ)
print(f"Descriptor ring @ phys {desc_ring_phys:016X} iova {desc_ring_iova:016X}")

apr.DR_HEAD = 0
apr.DR_TAIL = 0
apr.DR_SIZE = DESC_RING_SZ
apr.DR_ADDR_LO = desc_ring_iova & 0xFFFFFFFF
apr.DR_ADDR_HI = desc_ring_iova >> 32

apr.MODE = 0xd  # FIXME: dunno what this means

# MATRICES
apr.QUANT_LUMA_EHQ[0].val       = 0x802802
apr.QUANT_CHROMA_EHQ[0].val     = 0x804804
apr.QUANT_LUMA_EHQ[1].val       = 0x802802
apr.QUANT_CHROMA_EHQ[1].val     = 0x804804
apr.QUANT_LUMA_EHQ[2].val       = 0x802802
apr.QUANT_CHROMA_EHQ[2].val     = 0x804804
apr.QUANT_LUMA_EHQ[3].val       = 0x802802
apr.QUANT_CHROMA_EHQ[3].val     = 0x804804
apr.QUANT_LUMA_EHQ[4].val       = 0x802802
apr.QUANT_CHROMA_EHQ[4].val     = 0x804804
apr.QUANT_LUMA_EHQ[5].val       = 0x802802
apr.QUANT_CHROMA_EHQ[5].val     = 0x804804
apr.QUANT_LUMA_EHQ[6].val       = 0x802802
apr.QUANT_CHROMA_EHQ[6].val     = 0x804804
apr.QUANT_LUMA_EHQ[7].val       = 0x802802
apr.QUANT_CHROMA_EHQ[7].val     = 0x804804
apr.QUANT_LUMA_EHQ[8].val       = 0x802802
apr.QUANT_CHROMA_EHQ[8].val     = 0x804804
apr.QUANT_LUMA_EHQ[9].val       = 0x802802
apr.QUANT_CHROMA_EHQ[9].val     = 0x804804
apr.QUANT_LUMA_EHQ[10].val      = 0x802802
apr.QUANT_CHROMA_EHQ[10].val    = 0x804804
apr.QUANT_LUMA_EHQ[11].val      = 0x802802
apr.QUANT_CHROMA_EHQ[11].val    = 0x804804
apr.QUANT_LUMA_EHQ[12].val      = 0x802802
apr.QUANT_CHROMA_EHQ[12].val    = 0x804804
apr.QUANT_LUMA_EHQ[13].val      = 0x802802
apr.QUANT_CHROMA_EHQ[13].val    = 0x804804
apr.QUANT_LUMA_EHQ[14].val      = 0x802802
apr.QUANT_CHROMA_EHQ[14].val    = 0x804804
apr.QUANT_LUMA_EHQ[15].val      = 0x803802
apr.QUANT_CHROMA_EHQ[15].val    = 0x805804
apr.QUANT_LUMA_EHQ[16].val      = 0x802802
apr.QUANT_CHROMA_EHQ[16].val    = 0x804804
apr.QUANT_LUMA_EHQ[17].val      = 0x802802
apr.QUANT_CHROMA_EHQ[17].val    = 0x804804
apr.QUANT_LUMA_EHQ[18].val      = 0x802802
apr.QUANT_CHROMA_EHQ[18].val    = 0x804804
apr.QUANT_LUMA_EHQ[19].val      = 0x803803
apr.QUANT_CHROMA_EHQ[19].val    = 0x805805
apr.QUANT_LUMA_EHQ[20].val      = 0x802802
apr.QUANT_CHROMA_EHQ[20].val    = 0x804804
apr.QUANT_LUMA_EHQ[21].val      = 0x802802
apr.QUANT_CHROMA_EHQ[21].val    = 0x804804
apr.QUANT_LUMA_EHQ[22].val      = 0x803802
apr.QUANT_CHROMA_EHQ[22].val    = 0x805804
apr.QUANT_LUMA_EHQ[23].val      = 0x803803
apr.QUANT_CHROMA_EHQ[23].val    = 0x806805
apr.QUANT_LUMA_EHQ[24].val      = 0x802802
apr.QUANT_CHROMA_EHQ[24].val    = 0x804804
apr.QUANT_LUMA_EHQ[25].val      = 0x802802
apr.QUANT_CHROMA_EHQ[25].val    = 0x804804
apr.QUANT_LUMA_EHQ[26].val      = 0x803803
apr.QUANT_CHROMA_EHQ[26].val    = 0x805805
apr.QUANT_LUMA_EHQ[27].val      = 0x804803
apr.QUANT_CHROMA_EHQ[27].val    = 0x807806
apr.QUANT_LUMA_EHQ[28].val      = 0x802802
apr.QUANT_CHROMA_EHQ[28].val    = 0x804804
apr.QUANT_LUMA_EHQ[29].val      = 0x802802
apr.QUANT_CHROMA_EHQ[29].val    = 0x804804
apr.QUANT_LUMA_EHQ[30].val      = 0x803803
apr.QUANT_CHROMA_EHQ[30].val    = 0x806805
apr.QUANT_LUMA_EHQ[31].val      = 0x804804
apr.QUANT_CHROMA_EHQ[31].val    = 0x807807

apr.QUANT_LUMA_HQ[0].val        = 0x804804
apr.QUANT_CHROMA_HQ[0].val      = 0x804804
apr.QUANT_LUMA_HQ[1].val        = 0x804804
apr.QUANT_CHROMA_HQ[1].val      = 0x804804
apr.QUANT_LUMA_HQ[2].val        = 0x804804
apr.QUANT_CHROMA_HQ[2].val      = 0x804804
apr.QUANT_LUMA_HQ[3].val        = 0x804804
apr.QUANT_CHROMA_HQ[3].val      = 0x804804
apr.QUANT_LUMA_HQ[4].val        = 0x804804
apr.QUANT_CHROMA_HQ[4].val      = 0x804804
apr.QUANT_LUMA_HQ[5].val        = 0x804804
apr.QUANT_CHROMA_HQ[5].val      = 0x804804
apr.QUANT_LUMA_HQ[6].val        = 0x804804
apr.QUANT_CHROMA_HQ[6].val      = 0x804804
apr.QUANT_LUMA_HQ[7].val        = 0x804804
apr.QUANT_CHROMA_HQ[7].val      = 0x804804
apr.QUANT_LUMA_HQ[8].val        = 0x804804
apr.QUANT_CHROMA_HQ[8].val      = 0x804804
apr.QUANT_LUMA_HQ[9].val        = 0x804804
apr.QUANT_CHROMA_HQ[9].val      = 0x804804
apr.QUANT_LUMA_HQ[10].val       = 0x804804
apr.QUANT_CHROMA_HQ[10].val     = 0x804804
apr.QUANT_LUMA_HQ[11].val       = 0x804804
apr.QUANT_CHROMA_HQ[11].val     = 0x804804
apr.QUANT_LUMA_HQ[12].val       = 0x804804
apr.QUANT_CHROMA_HQ[12].val     = 0x804804
apr.QUANT_LUMA_HQ[13].val       = 0x804804
apr.QUANT_CHROMA_HQ[13].val     = 0x804804
apr.QUANT_LUMA_HQ[14].val       = 0x804804
apr.QUANT_CHROMA_HQ[14].val     = 0x804804
apr.QUANT_LUMA_HQ[15].val       = 0x805804
apr.QUANT_CHROMA_HQ[15].val     = 0x805804
apr.QUANT_LUMA_HQ[16].val       = 0x804804
apr.QUANT_CHROMA_HQ[16].val     = 0x804804
apr.QUANT_LUMA_HQ[17].val       = 0x804804
apr.QUANT_CHROMA_HQ[17].val     = 0x804804
apr.QUANT_LUMA_HQ[18].val       = 0x804804
apr.QUANT_CHROMA_HQ[18].val     = 0x804804
apr.QUANT_LUMA_HQ[19].val       = 0x805805
apr.QUANT_CHROMA_HQ[19].val     = 0x805805
apr.QUANT_LUMA_HQ[20].val       = 0x804804
apr.QUANT_CHROMA_HQ[20].val     = 0x804804
apr.QUANT_LUMA_HQ[21].val       = 0x804804
apr.QUANT_CHROMA_HQ[21].val     = 0x804804
apr.QUANT_LUMA_HQ[22].val       = 0x805804
apr.QUANT_CHROMA_HQ[22].val     = 0x805804
apr.QUANT_LUMA_HQ[23].val       = 0x806805
apr.QUANT_CHROMA_HQ[23].val     = 0x806805
apr.QUANT_LUMA_HQ[24].val       = 0x804804
apr.QUANT_CHROMA_HQ[24].val     = 0x804804
apr.QUANT_LUMA_HQ[25].val       = 0x804804
apr.QUANT_CHROMA_HQ[25].val     = 0x804804
apr.QUANT_LUMA_HQ[26].val       = 0x805805
apr.QUANT_CHROMA_HQ[26].val     = 0x805805
apr.QUANT_LUMA_HQ[27].val       = 0x807806
apr.QUANT_CHROMA_HQ[27].val     = 0x807806
apr.QUANT_LUMA_HQ[28].val       = 0x804804
apr.QUANT_CHROMA_HQ[28].val     = 0x804804
apr.QUANT_LUMA_HQ[29].val       = 0x804804
apr.QUANT_CHROMA_HQ[29].val     = 0x804804
apr.QUANT_LUMA_HQ[30].val       = 0x806805
apr.QUANT_CHROMA_HQ[30].val     = 0x806805
apr.QUANT_LUMA_HQ[31].val       = 0x807807
apr.QUANT_CHROMA_HQ[31].val     = 0x807807

apr.QUANT_LUMA_NQ[0].val        = 0x804804
apr.QUANT_CHROMA_NQ[0].val      = 0x804804
apr.QUANT_LUMA_NQ[1].val        = 0x805805
apr.QUANT_CHROMA_NQ[1].val      = 0x805805
apr.QUANT_LUMA_NQ[2].val        = 0x807806
apr.QUANT_CHROMA_NQ[2].val      = 0x807806
apr.QUANT_LUMA_NQ[3].val        = 0x809807
apr.QUANT_CHROMA_NQ[3].val      = 0x809807
apr.QUANT_LUMA_NQ[4].val        = 0x804804
apr.QUANT_CHROMA_NQ[4].val      = 0x804804
apr.QUANT_LUMA_NQ[5].val        = 0x806805
apr.QUANT_CHROMA_NQ[5].val      = 0x806805
apr.QUANT_LUMA_NQ[6].val        = 0x807807
apr.QUANT_CHROMA_NQ[6].val      = 0x807807
apr.QUANT_LUMA_NQ[7].val        = 0x809809
apr.QUANT_CHROMA_NQ[7].val      = 0x809809
apr.QUANT_LUMA_NQ[8].val        = 0x805805
apr.QUANT_CHROMA_NQ[8].val      = 0x805805
apr.QUANT_LUMA_NQ[9].val        = 0x807806
apr.QUANT_CHROMA_NQ[9].val      = 0x807806
apr.QUANT_LUMA_NQ[10].val       = 0x809807
apr.QUANT_CHROMA_NQ[10].val     = 0x809807
apr.QUANT_LUMA_NQ[11].val       = 0x80a809
apr.QUANT_CHROMA_NQ[11].val     = 0x80a809
apr.QUANT_LUMA_NQ[12].val       = 0x805805
apr.QUANT_CHROMA_NQ[12].val     = 0x805805
apr.QUANT_LUMA_NQ[13].val       = 0x807806
apr.QUANT_CHROMA_NQ[13].val     = 0x807806
apr.QUANT_LUMA_NQ[14].val       = 0x809807
apr.QUANT_CHROMA_NQ[14].val     = 0x809807
apr.QUANT_LUMA_NQ[15].val       = 0x80a809
apr.QUANT_CHROMA_NQ[15].val     = 0x80a809
apr.QUANT_LUMA_NQ[16].val       = 0x806805
apr.QUANT_CHROMA_NQ[16].val     = 0x806805
apr.QUANT_LUMA_NQ[17].val       = 0x807807
apr.QUANT_CHROMA_NQ[17].val     = 0x807807
apr.QUANT_LUMA_NQ[18].val       = 0x809808
apr.QUANT_CHROMA_NQ[18].val     = 0x809808
apr.QUANT_LUMA_NQ[19].val       = 0x80c80a
apr.QUANT_CHROMA_NQ[19].val     = 0x80c80a
apr.QUANT_LUMA_NQ[20].val       = 0x807806
apr.QUANT_CHROMA_NQ[20].val     = 0x807806
apr.QUANT_LUMA_NQ[21].val       = 0x808807
apr.QUANT_CHROMA_NQ[21].val     = 0x808807
apr.QUANT_LUMA_NQ[22].val       = 0x80a809
apr.QUANT_CHROMA_NQ[22].val     = 0x80a809
apr.QUANT_LUMA_NQ[23].val       = 0x80f80c
apr.QUANT_CHROMA_NQ[23].val     = 0x80f80c
apr.QUANT_LUMA_NQ[24].val       = 0x807806
apr.QUANT_CHROMA_NQ[24].val     = 0x807806
apr.QUANT_LUMA_NQ[25].val       = 0x809807
apr.QUANT_CHROMA_NQ[25].val     = 0x809807
apr.QUANT_LUMA_NQ[26].val       = 0x80b80a
apr.QUANT_CHROMA_NQ[26].val     = 0x80b80a
apr.QUANT_LUMA_NQ[27].val       = 0x81180e
apr.QUANT_CHROMA_NQ[27].val     = 0x81180e
apr.QUANT_LUMA_NQ[28].val       = 0x807807
apr.QUANT_CHROMA_NQ[28].val     = 0x807807
apr.QUANT_LUMA_NQ[29].val       = 0x80a809
apr.QUANT_CHROMA_NQ[29].val     = 0x80a809
apr.QUANT_LUMA_NQ[30].val       = 0x80e80b
apr.QUANT_CHROMA_NQ[30].val     = 0x80e80b
apr.QUANT_LUMA_NQ[31].val       = 0x815811
apr.QUANT_CHROMA_NQ[31].val     = 0x815811

apr.QUANT_LUMA_LT[0].val        = 0x805804
apr.QUANT_CHROMA_LT[0].val      = 0x805804
apr.QUANT_LUMA_LT[1].val        = 0x807806
apr.QUANT_CHROMA_LT[1].val      = 0x807806
apr.QUANT_LUMA_LT[2].val        = 0x80b809
apr.QUANT_CHROMA_LT[2].val      = 0x80b809
apr.QUANT_LUMA_LT[3].val        = 0x80f80d
apr.QUANT_CHROMA_LT[3].val      = 0x80f80d
apr.QUANT_LUMA_LT[4].val        = 0x805805
apr.QUANT_CHROMA_LT[4].val      = 0x805805
apr.QUANT_LUMA_LT[5].val        = 0x808807
apr.QUANT_CHROMA_LT[5].val      = 0x808807
apr.QUANT_LUMA_LT[6].val        = 0x80d80b
apr.QUANT_CHROMA_LT[6].val      = 0x80d80b
apr.QUANT_LUMA_LT[7].val        = 0x81180f
apr.QUANT_CHROMA_LT[7].val      = 0x81180f
apr.QUANT_LUMA_LT[8].val        = 0x807806
apr.QUANT_CHROMA_LT[8].val      = 0x807806
apr.QUANT_LUMA_LT[9].val        = 0x80b809
apr.QUANT_CHROMA_LT[9].val      = 0x80b809
apr.QUANT_LUMA_LT[10].val       = 0x80f80d
apr.QUANT_CHROMA_LT[10].val     = 0x80f80d
apr.QUANT_LUMA_LT[11].val       = 0x81180f
apr.QUANT_CHROMA_LT[11].val     = 0x81180f
apr.QUANT_LUMA_LT[12].val       = 0x807807
apr.QUANT_CHROMA_LT[12].val     = 0x807807
apr.QUANT_LUMA_LT[13].val       = 0x80b809
apr.QUANT_CHROMA_LT[13].val     = 0x80b809
apr.QUANT_LUMA_LT[14].val       = 0x80f80d
apr.QUANT_CHROMA_LT[14].val     = 0x80f80d
apr.QUANT_LUMA_LT[15].val       = 0x813811
apr.QUANT_CHROMA_LT[15].val     = 0x813811
apr.QUANT_LUMA_LT[16].val       = 0x809807
apr.QUANT_CHROMA_LT[16].val     = 0x809807
apr.QUANT_LUMA_LT[17].val       = 0x80d80b
apr.QUANT_CHROMA_LT[17].val     = 0x80d80b
apr.QUANT_LUMA_LT[18].val       = 0x81080e
apr.QUANT_CHROMA_LT[18].val     = 0x81080e
apr.QUANT_LUMA_LT[19].val       = 0x817813
apr.QUANT_CHROMA_LT[19].val     = 0x817813
apr.QUANT_LUMA_LT[20].val       = 0x80b809
apr.QUANT_CHROMA_LT[20].val     = 0x80b809
apr.QUANT_LUMA_LT[21].val       = 0x80e80d
apr.QUANT_CHROMA_LT[21].val     = 0x80e80d
apr.QUANT_LUMA_LT[22].val       = 0x813810
apr.QUANT_CHROMA_LT[22].val     = 0x813810
apr.QUANT_LUMA_LT[23].val       = 0x81d817
apr.QUANT_CHROMA_LT[23].val     = 0x81d817
apr.QUANT_LUMA_LT[24].val       = 0x80b809
apr.QUANT_CHROMA_LT[24].val     = 0x80b809
apr.QUANT_LUMA_LT[25].val       = 0x80f80d
apr.QUANT_CHROMA_LT[25].val     = 0x80f80d
apr.QUANT_LUMA_LT[26].val       = 0x815811
apr.QUANT_CHROMA_LT[26].val     = 0x815811
apr.QUANT_LUMA_LT[27].val       = 0x82381c
apr.QUANT_CHROMA_LT[27].val     = 0x82381c
apr.QUANT_LUMA_LT[28].val       = 0x80d80b
apr.QUANT_CHROMA_LT[28].val     = 0x80d80b
apr.QUANT_LUMA_LT[29].val       = 0x811810
apr.QUANT_CHROMA_LT[29].val     = 0x811810
apr.QUANT_LUMA_LT[30].val       = 0x81c815
apr.QUANT_CHROMA_LT[30].val     = 0x81c815
apr.QUANT_LUMA_LT[31].val       = 0x829823
apr.QUANT_CHROMA_LT[31].val     = 0x829823

apr.QUANT_LUMA_PROXY[0].val     = 0x807804
apr.QUANT_CHROMA_PROXY[0].val   = 0x807804
apr.QUANT_LUMA_PROXY[1].val     = 0x80b809
apr.QUANT_CHROMA_PROXY[1].val   = 0x80b809
apr.QUANT_LUMA_PROXY[2].val     = 0x80e80d
apr.QUANT_CHROMA_PROXY[2].val   = 0x80e80d
apr.QUANT_LUMA_PROXY[3].val     = 0xfff83f
apr.QUANT_CHROMA_PROXY[3].val   = 0xfff83f
apr.QUANT_LUMA_PROXY[4].val     = 0x807807
apr.QUANT_CHROMA_PROXY[4].val   = 0x807807
apr.QUANT_LUMA_PROXY[5].val     = 0x80c80b
apr.QUANT_CHROMA_PROXY[5].val   = 0x80c80b
apr.QUANT_LUMA_PROXY[6].val     = 0x83f80e
apr.QUANT_CHROMA_PROXY[6].val   = 0x83f80e
apr.QUANT_LUMA_PROXY[7].val     = 0xffffff
apr.QUANT_CHROMA_PROXY[7].val   = 0xffffff
apr.QUANT_LUMA_PROXY[8].val     = 0x80b809
apr.QUANT_CHROMA_PROXY[8].val   = 0x80b809
apr.QUANT_LUMA_PROXY[9].val     = 0x80e80d
apr.QUANT_CHROMA_PROXY[9].val   = 0x80e80d
apr.QUANT_LUMA_PROXY[10].val    = 0xfff83f
apr.QUANT_CHROMA_PROXY[10].val  = 0xfff83f
apr.QUANT_LUMA_PROXY[11].val    = 0xffffff
apr.QUANT_CHROMA_PROXY[11].val  = 0xffffff
apr.QUANT_LUMA_PROXY[12].val    = 0x80b80b
apr.QUANT_CHROMA_PROXY[12].val  = 0x80b80b
apr.QUANT_LUMA_PROXY[13].val    = 0x80e80d
apr.QUANT_CHROMA_PROXY[13].val  = 0x80e80d
apr.QUANT_LUMA_PROXY[14].val    = 0xffffff
apr.QUANT_CHROMA_PROXY[14].val  = 0xffffff
apr.QUANT_LUMA_PROXY[15].val    = 0xffffff
apr.QUANT_CHROMA_PROXY[15].val  = 0xffffff
apr.QUANT_LUMA_PROXY[16].val    = 0x80d80b
apr.QUANT_CHROMA_PROXY[16].val  = 0x80d80b
apr.QUANT_LUMA_PROXY[17].val    = 0xfff80e
apr.QUANT_CHROMA_PROXY[17].val  = 0xfff80e
apr.QUANT_LUMA_PROXY[18].val    = 0xffffff
apr.QUANT_CHROMA_PROXY[18].val  = 0xffffff
apr.QUANT_LUMA_PROXY[19].val    = 0xffffff
apr.QUANT_CHROMA_PROXY[19].val  = 0xffffff
apr.QUANT_LUMA_PROXY[20].val    = 0x80e80d
apr.QUANT_CHROMA_PROXY[20].val  = 0x80e80d
apr.QUANT_LUMA_PROXY[21].val    = 0xffffff
apr.QUANT_CHROMA_PROXY[21].val  = 0xffffff
apr.QUANT_LUMA_PROXY[22].val    = 0xffffff
apr.QUANT_CHROMA_PROXY[22].val  = 0xffffff
apr.QUANT_LUMA_PROXY[23].val    = 0xffffff
apr.QUANT_CHROMA_PROXY[23].val  = 0xffffff
apr.QUANT_LUMA_PROXY[24].val    = 0xfff80d
apr.QUANT_CHROMA_PROXY[24].val  = 0xfff80d
apr.QUANT_LUMA_PROXY[25].val    = 0xffffff
apr.QUANT_CHROMA_PROXY[25].val  = 0xffffff
apr.QUANT_LUMA_PROXY[26].val    = 0xffffff
apr.QUANT_CHROMA_PROXY[26].val  = 0xffffff
apr.QUANT_LUMA_PROXY[27].val    = 0xffffff
apr.QUANT_CHROMA_PROXY[27].val  = 0xffffff
apr.QUANT_LUMA_PROXY[28].val    = 0xffffff
apr.QUANT_CHROMA_PROXY[28].val  = 0xffffff
apr.QUANT_LUMA_PROXY[29].val    = 0xffffff
apr.QUANT_CHROMA_PROXY[29].val  = 0xffffff
apr.QUANT_LUMA_PROXY[30].val    = 0xffffff
apr.QUANT_CHROMA_PROXY[30].val  = 0xffffff
apr.QUANT_LUMA_PROXY[31].val    = 0xffffff
apr.QUANT_CHROMA_PROXY[31].val  = 0xffffff

apr.DC_QUANT_SCALE[0].val       = 0x401
apr.DC_QUANT_SCALE[1].val       = 0x803
apr.DC_QUANT_SCALE[2].val       = 0xc05
apr.DC_QUANT_SCALE[3].val       = 0x1007
apr.DC_QUANT_SCALE[4].val       = 0x1409
apr.DC_QUANT_SCALE[5].val       = 0x180b
apr.DC_QUANT_SCALE[6].val       = 0x1c0d
apr.DC_QUANT_SCALE[7].val       = 0x200f
apr.DC_QUANT_SCALE[8].val       = 0x2411
apr.DC_QUANT_SCALE[9].val       = 0x2813
apr.DC_QUANT_SCALE[10].val      = 0x2c15
apr.DC_QUANT_SCALE[11].val      = 0x3017
apr.DC_QUANT_SCALE[12].val      = 0x3419
apr.DC_QUANT_SCALE[13].val      = 0x381b
apr.DC_QUANT_SCALE[14].val      = 0x3c1d
apr.DC_QUANT_SCALE[15].val      = 0x401f
apr.DC_QUANT_SCALE[16].val      = 0x4421
apr.DC_QUANT_SCALE[17].val      = 0x4823
apr.DC_QUANT_SCALE[18].val      = 0x4c25
apr.DC_QUANT_SCALE[19].val      = 0x5027
apr.DC_QUANT_SCALE[20].val      = 0x5429
apr.DC_QUANT_SCALE[21].val      = 0x582b
apr.DC_QUANT_SCALE[22].val      = 0x5c2d
apr.DC_QUANT_SCALE[23].val      = 0x602f
apr.DC_QUANT_SCALE[24].val      = 0x6431
apr.DC_QUANT_SCALE[25].val      = 0x6833
apr.DC_QUANT_SCALE[26].val      = 0x6c35
apr.DC_QUANT_SCALE[27].val      = 0x7037
apr.DC_QUANT_SCALE[28].val      = 0x7439
apr.DC_QUANT_SCALE[29].val      = 0x783b
apr.DC_QUANT_SCALE[30].val      = 0x7c3d
apr.DC_QUANT_SCALE[31].val      = 0x803f
apr.DC_QUANT_SCALE[32].val      = 0x8441
apr.DC_QUANT_SCALE[33].val      = 0x8843
apr.DC_QUANT_SCALE[34].val      = 0x8c45
apr.DC_QUANT_SCALE[35].val      = 0x9047
apr.DC_QUANT_SCALE[36].val      = 0x9449
apr.DC_QUANT_SCALE[37].val      = 0x984b
apr.DC_QUANT_SCALE[38].val      = 0x9c4d
apr.DC_QUANT_SCALE[39].val      = 0xa04f
apr.DC_QUANT_SCALE[40].val      = 0xa451
apr.DC_QUANT_SCALE[41].val      = 0xa853
apr.DC_QUANT_SCALE[42].val      = 0xac55
apr.DC_QUANT_SCALE[43].val      = 0xb057
apr.DC_QUANT_SCALE[44].val      = 0xb459
apr.DC_QUANT_SCALE[45].val      = 0xb85b
apr.DC_QUANT_SCALE[46].val      = 0xbc5d
apr.DC_QUANT_SCALE[47].val      = 0xc05f
apr.DC_QUANT_SCALE[48].val      = 0xc461
apr.DC_QUANT_SCALE[49].val      = 0xc863
apr.DC_QUANT_SCALE[50].val      = 0xcc65
apr.DC_QUANT_SCALE[51].val      = 0xd067
apr.DC_QUANT_SCALE[52].val      = 0xd469
apr.DC_QUANT_SCALE[53].val      = 0xd86b
apr.DC_QUANT_SCALE[54].val      = 0xdc6d
apr.DC_QUANT_SCALE[55].val      = 0xe06f
apr.DC_QUANT_SCALE[56].val      = 0xe471
apr.DC_QUANT_SCALE[57].val      = 0xe873
apr.DC_QUANT_SCALE[58].val      = 0xec75
apr.DC_QUANT_SCALE[59].val      = 0xf077
apr.DC_QUANT_SCALE[60].val      = 0xf479
apr.DC_QUANT_SCALE[61].val      = 0xf87b
apr.DC_QUANT_SCALE[62].val      = 0xfc7d
apr.DC_QUANT_SCALE[63].val      = 0x1007f
apr.DC_QUANT_SCALE[64].val      = 0x11084
apr.DC_QUANT_SCALE[65].val      = 0x1208c
apr.DC_QUANT_SCALE[66].val      = 0x13094
apr.DC_QUANT_SCALE[67].val      = 0x1409c
apr.DC_QUANT_SCALE[68].val      = 0x150a4
apr.DC_QUANT_SCALE[69].val      = 0x160ac
apr.DC_QUANT_SCALE[70].val      = 0x170b4
apr.DC_QUANT_SCALE[71].val      = 0x180bc
apr.DC_QUANT_SCALE[72].val      = 0x190c4
apr.DC_QUANT_SCALE[73].val      = 0x1a0cc
apr.DC_QUANT_SCALE[74].val      = 0x1b0d4
apr.DC_QUANT_SCALE[75].val      = 0x1c0dc
apr.DC_QUANT_SCALE[76].val      = 0x1d0e4
apr.DC_QUANT_SCALE[77].val      = 0x1e0ec
apr.DC_QUANT_SCALE[78].val      = 0x1f0f4
apr.DC_QUANT_SCALE[79].val      = 0x200fc
apr.DC_QUANT_SCALE[80].val      = 0x21104
apr.DC_QUANT_SCALE[81].val      = 0x2210c
apr.DC_QUANT_SCALE[82].val      = 0x23114
apr.DC_QUANT_SCALE[83].val      = 0x2411c
apr.DC_QUANT_SCALE[84].val      = 0x25124
apr.DC_QUANT_SCALE[85].val      = 0x2612c
apr.DC_QUANT_SCALE[86].val      = 0x27134
apr.DC_QUANT_SCALE[87].val      = 0x2813c
apr.DC_QUANT_SCALE[88].val      = 0x29144
apr.DC_QUANT_SCALE[89].val      = 0x2a14c
apr.DC_QUANT_SCALE[90].val      = 0x2b154
apr.DC_QUANT_SCALE[91].val      = 0x2c15c
apr.DC_QUANT_SCALE[92].val      = 0x2d164
apr.DC_QUANT_SCALE[93].val      = 0x2e16c
apr.DC_QUANT_SCALE[94].val      = 0x2f174
apr.DC_QUANT_SCALE[95].val      = 0x3017c
apr.DC_QUANT_SCALE[96].val      = 0x31184
apr.DC_QUANT_SCALE[97].val      = 0x3218c
apr.DC_QUANT_SCALE[98].val      = 0x33194
apr.DC_QUANT_SCALE[99].val      = 0x3419c
apr.DC_QUANT_SCALE[100].val     = 0x351a4
apr.DC_QUANT_SCALE[101].val     = 0x361ac
apr.DC_QUANT_SCALE[102].val     = 0x371b4
apr.DC_QUANT_SCALE[103].val     = 0x381bc
apr.DC_QUANT_SCALE[104].val     = 0x391c4
apr.DC_QUANT_SCALE[105].val     = 0x3a1cc
apr.DC_QUANT_SCALE[106].val     = 0x3b1d4
apr.DC_QUANT_SCALE[107].val     = 0x3c1dc
apr.DC_QUANT_SCALE[108].val     = 0x3d1e4
apr.DC_QUANT_SCALE[109].val     = 0x3e1ec
apr.DC_QUANT_SCALE[110].val     = 0x3f1f4
apr.DC_QUANT_SCALE[111].val     = 0x1fc
print("Set matrices")

# dunno how this gets calculated
OUT_SZ = 0x1000000
out_buf_phys = u.heap.memalign(0x4000, OUT_SZ)
iface.writemem(out_buf_phys, b'\xAA' * OUT_SZ)
out_buf_iova = dart.iomap(0, out_buf_phys, OUT_SZ)
print(f"Output buffer @ phys {out_buf_phys:016X} iova {out_buf_iova:016X}")

IN_SZ_LUMA = align_up(im_W*im_H)
in_buf_luma_phys = u.heap.memalign(0x4000, IN_SZ_LUMA)
iface.writemem(in_buf_luma_phys, image_data_luma + b'\xaa' * (IN_SZ_LUMA - len(image_data_luma)))
in_buf_luma_iova = dart.iomap(0, in_buf_luma_phys, IN_SZ_LUMA)
print(f"Input buffer luma @ phys {in_buf_luma_phys:016X} iova {in_buf_luma_iova:016X}")
IN_SZ_CHROMA = align_up(im_W*2*im_H)
in_buf_chroma_phys = u.heap.memalign(0x4000, IN_SZ_CHROMA)
iface.writemem(in_buf_chroma_phys, image_data_chroma + b'\xaa' * (IN_SZ_CHROMA - len(image_data_chroma)))
in_buf_chroma_iova = dart.iomap(0, in_buf_chroma_phys, IN_SZ_CHROMA)
print(f"Input buffer chroma @ phys {in_buf_chroma_phys:016X} iova {in_buf_chroma_iova:016X}")
dart.dump_all()

desc = EncodeNotRawDescriptor(
    flags=0x373c,
    flags2=0,
    output_iova=out_buf_iova,
    max_out_sz=OUT_SZ,
    offset_x=0,
    offset_y=0,
    # this is the important set
    pix_surface_w_2_=im_W,
    pix_surface_h_2_=im_H,
    # changing this doesn't seem to break anything
    pix_surface_w=im_W,
    pix_surface_h=im_H,
    # XXX how does the div work exactly? it's different in "tiled" mode
    luma_stride=divroundup(im_W, 64),
    chroma_stride=divroundup(im_W*2, 64),
    alpha_stride=divroundup(im_W, 64),
    unk_pad_0x26_=b'\x00\x00',

    luma_iova=in_buf_luma_iova,
    pix_plane0_tileheader_thing_=0,
    chroma_iova=in_buf_chroma_iova,
    pix_plane1_tileheader_thing_=0,
    alpha_iova=in_buf_luma_iova,
    pix_plane2_tileheader_thing_=0,

    # changing this does add extra 0 bytes
    frame_header_sz=bswp16(0x94),
    unk_pad_0x5a_=b'\x00',
    bitstream_version=0,
    encoder_identifier=0xcafeface,
    # cannot change arbitrarily, will break
    pix_surface_w_byteswap_=bswp16(im_W),
    pix_surface_h_byteswap_=bswp16(im_H),
    # seemingly can change arbitrarily
    chroma_format_interlace_mode=0xc0,
    aspect_ratio_frame_rate=0,
    color_primaries=2,
    transfer_characteristic=2,
    matrix_coefficients=1,
    alpha_channel_type=1,
    # tables will still be output even if bits not set here
    frame_hdr_reserved14=b'\x00\x03',
    unk_pad_0x6c_=b'\x00' * 128,
    deprecated_number_of_slices=0,
    # this one affects the encoding not just the header
    log2_desired_slice_size_in_mb=0x30,
    quantization_index=0x2,

    # this impacts the quality somehow, not quite understood
    # might be a target bitrate
    unk_0xf0_=0xffff,
    unk_0xf2_=0xffff,

    # none of this stuff is understood, and it never seems to change
    unk_0xf4_=0x8000402015100c0c,
    unk_0xfc_=0x2c8080,
    unk_0x100_0_=0x880080,
    unk_0x100_1_=0x4e00c5,
    unk_0x100_2_=0x9000d0,
    unk_0x100_3_=0x200122,
    unk_0x110_0_=0x400200,  # looks like a quant table, but ??? not used ???
    unk_0x110_1_=0x400200,
    unk_0x110_2_=0x400200,
    unk_0x110_3_=0x400200,
    unk_0x110_4_=0x400200,
    unk_0x110_5_=0x400200,
    unk_0x110_6_=0x400200,
    unk_0x110_7_=0x400200,
    unk_0x110_8_=0x400200,
    unk_0x110_9_=0x400200,
    unk_0x110_10_=0x400200,
    unk_0x110_11_=0x400200,
    unk_0x110_12_=0x400200,
    unk_0x110_13_=0x400200,
    unk_0x110_14_=0x400200,
    unk_0x110_15_=0x400200,

    quant_table_sel=0x23,
    unk_pad_0x154_=b'\x00' * 44,
)
desc_bytes = struct.pack(ENCODE_NOT_RAW_STRUCT, *desc)
chexdump(desc_bytes)

iface.writemem(desc_ring_phys, desc_bytes)

# let's go
apr.DR_HEAD = len(desc_bytes)

start_time = time.time()
while apr.IRQ_STATUS.val == 0:
    if time.time() - start_time > 5:
        print("TIMED OUT!!!")
        break

print(f"Done, IRQ status is {apr.IRQ_STATUS}")
print(f"ST0 = {apr.ST0}")
print(f"ST1 = {apr.ST1}")
print(f"REG_0x1c = {apr.REG_0x1c}")
print(f"REG_0x3c = {apr.REG_0x3c}")
print(f"REG_0x44 = {apr.REG_0x44}")

print(f"DR_HEAD = {apr.DR_HEAD}")
print(f"DR_TAIL = {apr.DR_TAIL}")

print(f"unk REG_0x38 = {apr.REG_0x38}")
print(f"unk REG_0x40 = {apr.REG_0x40}")
print(f"unk REG_0x48 = {apr.REG_0x48}")
print(f"unk REG_0x50 = {apr.REG_0x50}")
print(f"unk REG_0x54 = {apr.REG_0x54}")

apr.IRQ_STATUS = apr.IRQ_STATUS.val

dr_memory_new = iface.readmem(desc_ring_phys, DESC_RING_SZ)
chexdump(dr_memory_new)

out_buf_new = iface.readmem(out_buf_phys, OUT_SZ)
with open('prores.bin', 'wb') as f:
    f.write(out_buf_new)

outlen = struct.unpack(">I", out_buf_new[:4])[0]
if outlen <= len(out_buf_new):
    with open('prores.mov', 'wb') as f:
        f.write(b'\x00\x00\x00\x14\x66\x74\x79\x70\x71\x74\x20\x20\x00\x00\x02\x00\x71\x74\x20\x20\x00\x00\x00\x08\x77\x69\x64\x65')
        f.write(struct.pack(">I", outlen + 8))
        f.write(b'mdat')
        f.write(out_buf_new[:outlen])
        f.write(b'\x00\x00\x03\x1e\x6d\x6f\x6f\x76\x00\x00\x00\x6c\x6d\x76\x68\x64\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x03\xe8\x00\x00\x00\x11\x00\x01\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x40\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x02\x00\x00\x02\x89\x74\x72\x61\x6b\x00\x00\x00\x5c\x74\x6b\x68\x64\x00\x00\x00\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x11\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x40\x00\x00\x00\x07\x80\x00\x00\x04\x38\x00\x00\x00\x00\x00\x24\x65\x64\x74\x73\x00\x00\x00\x1c\x65\x6c\x73\x74\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x11\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\x02\x01\x6d\x64\x69\x61\x00\x00\x00\x20\x6d\x64\x68\x64\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3c\x00\x00\x00\x01\x00\x7f\xff\x00\x00\x00\x00\x00\x2d\x68\x64\x6c\x72\x00\x00\x00\x00\x6d\x68\x6c\x72\x76\x69\x64\x65\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x0c\x56\x69\x64\x65\x6f\x48\x61\x6e\x64\x6c\x65\x72\x00\x00\x01\xac\x6d\x69\x6e\x66\x00\x00\x00\x14\x76\x6d\x68\x64\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x2c\x68\x64\x6c\x72\x00\x00\x00\x00\x64\x68\x6c\x72\x75\x72\x6c\x20\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x0b\x44\x61\x74\x61\x48\x61\x6e\x64\x6c\x65\x72\x00\x00\x00\x24\x64\x69\x6e\x66\x00\x00\x00\x1c\x64\x72\x65\x66\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x0c\x75\x72\x6c\x20\x00\x00\x00\x01\x00\x00\x01\x40\x73\x74\x62\x6c\x00\x00\x00\xdc\x73\x74\x73\x64\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\xcc\x61\x70\x63\x6e\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\x46\x46\x4d\x50\x00\x00\x02\x00\x00\x00\x02\x00\x07\x80\x04\x38\x00\x48\x00\x00\x00\x48\x00\x00\x00\x00\x00\x00\x00\x01\x1f\x4c\x61\x76\x63\x35\x39\x2e\x32\x35\x2e\x31\x30\x30\x20\x70\x72\x6f\x72\x65\x73\x5f\x76\x69\x64\x65\x6f\x74\x6f\x6f\x6c\x62\x00\x18\xff\xff\x00\x00\x00\x6c\x67\x6c\x62\x6c\x00\x00\x00\x64\x61\x70\x63\x6e\x00\x00\x00\x00\x00\x00\xff\xff\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x03\xff\x07\x80\x04\x38\x00\x48\x00\x00\x00\x48\x00\x00\x00\x00\x00\x00\x00\x01\x10\x41\x70\x70\x6c\x65\x20\x50\x72\x6f\x52\x65\x73\x20\x34\x32\x32\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x18\xff\xff\x00\x00\x00\x0a\x66\x69\x65\x6c\x01\x00\x00\x00\x00\x00\x00\x00\x00\x0a\x66\x69\x65\x6c\x01\x00\x00\x00\x00\x18\x73\x74\x74\x73\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x01\x00\x00\x01\x00\x00\x00\x00\x1c\x73\x74\x73\x63\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x01\x00\x00\x00\x01\x00\x00\x00\x01\x00\x00\x00\x14\x73\x74\x73\x7a\x00\x00\x00\x00')
        f.write(struct.pack(">I", outlen))
        f.write(b'\x00\x00\x00\x01\x00\x00\x00\x14\x73\x74\x63\x6f\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x24\x00\x00\x00\x21\x75\x64\x74\x61\x00\x00\x00\x19\xa9\x73\x77\x72\x00\x0d\x55\xc4\x4c\x61\x76\x66\x35\x39\x2e\x32\x30\x2e\x31\x30\x31')
# ffmpeg -i prores.mov prores-dec%d.png



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib
import time
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

# audio_capture.py -- capture audio on jack microphone input (on M1 macs with cs42l83)
#
# sample usage with sox: (recoding can be loud!)
#
#  ./audio_capture.py | sox -t raw -r 48000 -c 1 -e signed-int -b 32 -L - OUTPUT_FILE

from m1n1.setup import *
from m1n1.hw.dart import DART, DARTRegs
from m1n1.hw.i2c import I2C
from m1n1.hw.pmgr import PMGR
from m1n1.hw.nco import NCO
from m1n1.hw.admac import *
from m1n1.hw.mca import *

p.pmgr_adt_clocks_enable("/arm-io/i2c2")
p.pmgr_adt_clocks_enable("/arm-io/admac-sio")
p.pmgr_adt_clocks_enable("/arm-io/dart-sio")
p.pmgr_adt_clocks_enable("/arm-io/mca-switch")
p.pmgr_adt_clocks_enable("/arm-io/mca3")

# reset AUDIO_P
PS_AUDIO_P = PMGR(u).regs[0].PS4[10]
PS_AUDIO_P.set(DEV_DISABLE=1)
PS_AUDIO_P.set(RESET=1)
PS_AUDIO_P.set(RESET=0)
PS_AUDIO_P.set(DEV_DISABLE=0)

dart_base, _ = u.adt["/arm-io/dart-sio"].get_reg(0)
dart = DART(iface, DARTRegs(u, dart_base), util=u)
dart.initialize()

cl_no = 2

admac = ADMAC(u, "/arm-io/admac-sio", dart, debug=True)
dmachan = admac.chans[4*cl_no+1]
dmachan.buswidth = E_BUSWIDTH.W_32BIT
dmachan.framesize = E_FRAME.F_1_WORD

nco = NCO(u, "/arm-io/nco")
nco[cl_no].set_rate(6000000)
nco[cl_no].enable()

mca_switch1_base = u.adt["/arm-io/mca-switch"].get_reg(1)[0]
mca_cl_base = u.adt["/arm-io/mca-switch"].get_reg(0)[0] + 0x4000*cl_no
cl = MCACluster(u, mca_cl_base)

regs, serdes = cl.regs, cl.rxa

regs.SYNCGEN_STATUS.set(EN=0)
regs.SYNCGEN_MCLK_SEL.val =(1 + cl_no)
regs.SYNCGEN_HI_PERIOD.val = 0    # period minus one
regs.SYNCGEN_LO_PERIOD.val = 0x7b # period minus one

serdes.STATUS.set(EN=0)
serdes.CONF.set(
	NSLOTS=0,
	SLOT_WIDTH=E_SLOT_WIDTH.W_32BIT,
	BCLK_POL=1,
	UNK1=1, UNK2=1,
	SYNC_SEL=(1 + cl_no)
)
serdes.UNK1.val = 0x4

serdes.BITDELAY.val = 1

serdes.CHANMASK[0].val = 0xffff_ffff
serdes.CHANMASK[1].val = 0xffff_fffe

regs.PORT_ENABLES.set(CLOCK1=1, CLOCK2=1, DATA=0)
regs.PORT_CLK_SEL.set(SEL=(cl_no + 1))
regs.MCLK_STATUS.set(EN=1)
regs.SYNCGEN_STATUS.set(EN=1)

cs42l_addr = 0x48
i2c2 = I2C(u, "/arm-io/i2c2")
def cs42l_write(regaddr, val):
	i2c2.write_reg(cs42l_addr, 0x0, [regaddr >> 8])
	i2c2.write_reg(cs42l_addr, regaddr & 0xff, [val])

p.write32(0x23d1f002c, 0x76a02)
p.write32(0x23d1f002c, 0x76a03) # take jack codec out of reset

cs42l_write(0x1009, 0x0)  # FS_int = MCLK/250
cs42l_write(0x1101, 0x7a) # power on
cs42l_write(0x1103, 0x22) # power on ring sense
cs42l_write(0x1107, 0x1)  # SCLK present
cs42l_write(0x1121, 0xa6) # Headset Switch Control
cs42l_write(0x1129, 0x1)  # Headset Clamp Disable
cs42l_write(0x1205, 0x7c) # FSYNC period
cs42l_write(0x1207, 0x20) # ASP Clock Configuration
cs42l_write(0x1208, 0x12) # BITDELAY = 1
cs42l_write(0x120c, 0x1)  # SCLK_PREDIV = div-by-2
cs42l_write(0x150a, 0x55) # PLL
cs42l_write(0x151b, 0x1)  # PLL
cs42l_write(0x1501, 0x1)  # power on PLL
cs42l_write(0x1b70, 0xc3) # HSBIAS sense
cs42l_write(0x1b71, 0xe0) # v-- headset 
cs42l_write(0x1b73, 0xc0)
cs42l_write(0x1b74, 0x1f)
cs42l_write(0x1b75, 0xb6)
cs42l_write(0x1b76, 0x8f)
cs42l_write(0x1b79, 0x0)
cs42l_write(0x1b7a, 0xfc)
cs42l_write(0x1c03, 0xc0) # HSBIAS
cs42l_write(0x2506, 0xc)  # ASP TX samp. rate
cs42l_write(0x2609, 0x4c) # SRC output samp. rate
cs42l_write(0x2901, 0x1)  # ASP TX enable & size
cs42l_write(0x2902, 0x1)  # ASP TX channel enable

time.sleep(0.01)

cs42l_write(0x1201, 0x1) # transition to PLL clock

# drain garbled samples (why are they garbled? i am not sure)
time.sleep(0.5)

dmachan.submit(buflen=0x4000)
dmachan.enable()

p.write32(mca_switch1_base + 0x8000*cl_no, 0x24800)
serdes.STATUS.set(EN=1)

while True:
	while dmachan.can_submit():
		dmachan.submit(buflen=0x4000)
	sys.stdout.buffer.write(dmachan.poll())



# SPDX-License-Identifier: MIT
from ..utils import chexdump
from ..malloc import Heap
from construct.core import *
from ..fw.agx.channels import *
from ..fw.agx.cmdqueue import *
from ..fw.agx.microsequence import *
from ..hw.uat import MemoryAttr
from .object import *
import textwrap

class GPUEventManager:
    MAX_EVENTS = 128

    def __init__(self, agx):
        self.agx = agx

        self.event_count = 0
        self.free_events = set(range(self.MAX_EVENTS))
        self.events = [None] * self.MAX_EVENTS

    def allocate_event(self):
        if not self.free_events:
            raise Exception("No free events")
        ev_id = self.free_events.pop()

        ev = GPUEvent(ev_id)
        self.events[ev_id] = ev

        return ev

    def free_event(self, ev):
        self.events[ev.id] = None
        self.free_events.add(ev.id)

    def fired(self, flags):
        self.agx.log("= Events fired =")
        for i, v in enumerate(flags):
            for j in range(64):
                if v & (1 << j):
                    ev_id = i * 64 + j
                    ev = self.events[ev_id]
                    self.agx.log(f"Event fired: {ev_id}")
                    if ev is None:
                        raise Exception("Received spurious notification for event ID {ev}")
                    ev.fire()
                    self.event_count += 1

class GPUEvent:
    def __init__(self, ev_id):
        self.id = ev_id
        self.fired = False

    def fire(self):
        self.fired = True

    def rearm(self):
        self.fired = False



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT

import sys, pathlib, time
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

import atexit, sys

from m1n1.setup import *
from m1n1.constructutils import Ver
from m1n1.utils import *

Ver.set_version(u)

from m1n1.agx import AGX
from m1n1.agx.render import *

from m1n1.gpiola import GPIOLogicAnalyzer

analyzer_cpu = 1

p.pmgr_adt_clocks_enable("/arm-io/gfx-asc")
p.pmgr_adt_clocks_enable("/arm-io/sgx")
p.smp_start_secondaries()
p.mmu_init_secondary(analyzer_cpu)
iface.dev.timeout = 10

agx = AGX(u)

mon = RegMonitor(u, ascii=True, bufsize=0x8000000)
agx.mon = mon

sgx = agx.sgx_dev
#mon.add(sgx.gpu_region_base, sgx.gpu_region_size, "contexts")
#mon.add(sgx.gfx_shared_region_base, sgx.gfx_shared_region_size, "gfx-shared")
#mon.add(sgx.gfx_handoff_base, sgx.gfx_handoff_size, "gfx-handoff")

#mon.add(agx.initdasgx.gfx_handoff_base, sgx.gfx_handoff_size, "gfx-handoff")

atexit.register(p.reboot)
agx.start()

print("==========================================")
print("## After init")
print("==========================================")
mon.poll()
agx.poll_objects()

ctx = GPUContext(agx)
ctx.bind(1)

renderer = GPURenderer(ctx, 8, bm_slot=0, queue=0)
renderer2 = GPURenderer(ctx, 8, bm_slot=1, queue=1)

for q in (renderer2.wq_3d, renderer2.wq_ta):
    q.info.set_prio(0)
    q.info.push()

for q in (renderer.wq_3d, renderer.wq_ta):
    q.info.set_prio(2)
    q.info.push()

f = GPUFrame(ctx, sys.argv[1], track=False)

print("==========================================")
print("## Pre submit")
print("==========================================")

mon.poll()
agx.poll_objects()

print("==========================================")
print("## Submitting")
print("==========================================")

work = renderer.submit(f.cmdbuf)
work2 = renderer2.submit(f.cmdbuf)
workb = renderer.submit(f.cmdbuf)
work2b = renderer2.submit(f.cmdbuf)

print(work.wc_3d)
print(work.wc_ta)
print(work2.wc_3d)
print(work2.wc_ta)

print("==========================================")
print("## Submitted")
print("==========================================")

def t(addr):
    paddr = agx.uat.iotranslate(0, addr, 4)[0][0]
    if paddr is None:
        raise Exception(f"Failed to iotranslate {addr:#x}")
    return paddr

regs = {
    "ta_cmds":  t(agx.initdata.regionB.stats_ta.addrof("total_cmds")),
    "ta0_busy": t(agx.initdata.regionB.stats_ta.stats.queues[0].addrof("busy")),
    "ta0_unk4": t(agx.initdata.regionB.stats_ta.stats.queues[0].addrof("unk_4")),
    "ta0_cnt":  t(agx.initdata.regionB.stats_ta.stats.queues[0].addrof("cur_count")),
    "ta1_busy": t(agx.initdata.regionB.stats_ta.stats.queues[1].addrof("busy")),
    "ta1_unk4": t(agx.initdata.regionB.stats_ta.stats.queues[1].addrof("unk_4")),
    "ta1_cnt":  t(agx.initdata.regionB.stats_ta.stats.queues[1].addrof("cur_count")),
    "ta_ts":    t(agx.initdata.regionB.stats_ta.stats.addrof("unk_timestamp")),
    "3d_cmds":  t(agx.initdata.regionB.stats_3d.addrof("total_cmds")),
    "3d_tvb_oflws_1":   t(agx.initdata.regionB.stats_3d.stats.addrof("tvb_overflows_1")),
    "3d_tvb_oflws_2":   t(agx.initdata.regionB.stats_3d.stats.addrof("tvb_overflows_2")),
    "3d_cur_stamp_id":  t(agx.initdata.regionB.stats_3d.stats.addrof("cur_stamp_id")),
    "3d_ts":    t(agx.initdata.regionB.stats_3d.stats.addrof("unk_timestamp")),

    "bmctl_0":  t(agx.initdata.regionB.buffer_mgr_ctl._addr + 0),
    "bmctl_8":  t(agx.initdata.regionB.buffer_mgr_ctl._addr + 8),
    "2_bmctl_0":  t(agx.initdata.regionB.buffer_mgr_ctl._addr + 16),
    "2_bmctl_8":  t(agx.initdata.regionB.buffer_mgr_ctl._addr + 24),

    "bmi_gpuc":  t(renderer.buffer_mgr.info.addrof("gpu_counter")),
    "bmi_18":    t(renderer.buffer_mgr.info.addrof("unk_18")),
    "bmi_gpuc2": t(renderer.buffer_mgr.info.addrof("gpu_counter2")),

    "2_bmi_gpuc":  t(renderer2.buffer_mgr.info.addrof("gpu_counter")),
    "2_bmi_18":    t(renderer2.buffer_mgr.info.addrof("unk_18")),
    "2_bmi_gpuc2": t(renderer2.buffer_mgr.info.addrof("gpu_counter2")),

    "ctxdat_0": t(renderer.scheduler_context._addr + 0),
    "ctxdat_4": t(renderer.scheduler_context._addr + 4),
    "ctxdat_6": t(renderer.scheduler_context._addr + 6),
    "ctxdat_a": t(renderer.scheduler_context._addr + 0xa),
    "ctxdat_e": t(renderer.scheduler_context._addr + 0xe),
    "ctxdat_12": t(renderer.scheduler_context._addr + 0x12),
    "ctxdat_16": t(renderer.scheduler_context._addr + 0x16),
    "ctxdat_1a": t(renderer.scheduler_context._addr + 0x1a),
    "ctxdat_24": t(renderer.scheduler_context._addr + 0x24),
    "ctxdat_28": t(renderer.scheduler_context._addr + 0x28),
    "ctxdat_2c": t(renderer.scheduler_context._addr + 0x2c),

    "2_ctxdat_0": t(renderer2.scheduler_context._addr + 0),
    "2_ctxdat_4": t(renderer2.scheduler_context._addr + 4),
    "2_ctxdat_6": t(renderer2.scheduler_context._addr + 6),
    "2_ctxdat_a": t(renderer2.scheduler_context._addr + 0xa),
    "2_ctxdat_e": t(renderer2.scheduler_context._addr + 0xe),
    "2_ctxdat_12": t(renderer2.scheduler_context._addr + 0x12),
    "2_ctxdat_16": t(renderer2.scheduler_context._addr + 0x16),
    "2_ctxdat_1a": t(renderer2.scheduler_context._addr + 0x1a),
    "2_ctxdat_24": t(renderer2.scheduler_context._addr + 0x24),
    "2_ctxdat_28": t(renderer2.scheduler_context._addr + 0x28),
    "2_ctxdat_2c": t(renderer2.scheduler_context._addr + 0x2c),

    "evctl_ta":      t(renderer.event_control.addrof("has_ta")),
    "evctl_pta":     t(renderer.event_control.addrof("pstamp_ta")),
    "evctl_3d":      t(renderer.event_control.addrof("has_3d")),
    "evctl_p3d":     t(renderer.event_control.addrof("pstamp_3d")),
    "evctl_in_list": t(renderer.event_control.addrof("in_list")),
    "evctl_prev":    t(renderer.event_control.list_head.addrof("prev")),
    "evctl_next":    t(renderer.event_control.list_head.addrof("next")),

    "2_evctl_ta":    t(renderer2.event_control.addrof("has_ta")),
    "2_evctl_pta":   t(renderer2.event_control.addrof("pstamp_ta")),
    "2_evctl_3d":    t(renderer2.event_control.addrof("has_3d")),
    "2_evctl_p3d":   t(renderer2.event_control.addrof("pstamp_3d")),
    "2_evctl_in_list":t(renderer2.event_control.addrof("in_list")),
    "2_evctl_prev":  t(renderer2.event_control.list_head.addrof("prev")),
    "2_evctl_next":  t(renderer2.event_control.list_head.addrof("next")),

    "jl_first": t(renderer.job_list.addrof("first_job")),
    "jl_last":  t(renderer.job_list.addrof("last_head")),
    "jl_10":    t(renderer.job_list.addrof("unkptr_10")),

    "2_jl_first": t(renderer2.job_list.addrof("first_job")),
    "2_jl_last":  t(renderer2.job_list.addrof("last_head")),
    "2_jl_10":    t(renderer2.job_list.addrof("unkptr_10")),

    "3d_done":  t(renderer.wq_3d.info.pointers.addrof("gpu_doneptr")),
    "3d_rptr":  t(renderer.wq_3d.info.pointers.addrof("gpu_rptr")),
    "3d_rptr1": t(renderer.wq_3d.info.addrof("gpu_rptr1")),
    "3d_rptr2": t(renderer.wq_3d.info.addrof("gpu_rptr2")),
    "3d_rptr3": t(renderer.wq_3d.info.addrof("gpu_rptr3")),
    "3d_busy":  t(renderer.wq_3d.info.addrof("busy")),
    "3d_54":    t(renderer.wq_3d.info.addrof("unk_54")),
    "3d_58":    t(renderer.wq_3d.info.addrof("unk_58")),

    "2_3d_done":  t(renderer2.wq_3d.info.pointers.addrof("gpu_doneptr")),
    "2_3d_rptr":  t(renderer2.wq_3d.info.pointers.addrof("gpu_rptr")),
    "2_3d_busy":  t(renderer2.wq_3d.info.addrof("busy")),
    "2_3d_54":    t(renderer2.wq_3d.info.addrof("unk_54")),

    "ta_done":  t(renderer.wq_ta.info.pointers.addrof("gpu_doneptr")),
    "ta_rptr":  t(renderer.wq_ta.info.pointers.addrof("gpu_rptr")),
    "ta_rptr1": t(renderer.wq_ta.info.addrof("gpu_rptr1")),
    "ta_rptr2": t(renderer.wq_ta.info.addrof("gpu_rptr2")),
    "ta_rptr3": t(renderer.wq_ta.info.addrof("gpu_rptr3")),
    "ta_busy":  t(renderer.wq_ta.info.addrof("busy")),
    "ta_54":   t(renderer.wq_ta.info.addrof("unk_54")),
    "ta_58":   t(renderer.wq_ta.info.addrof("unk_58")),

    "2_ta_done":  t(renderer2.wq_ta.info.pointers.addrof("gpu_doneptr")),
    "2_ta_rptr":  t(renderer2.wq_ta.info.pointers.addrof("gpu_rptr")),
    "2_ta_busy":  t(renderer2.wq_ta.info.addrof("busy")),
    "2_ta_54":    t(renderer2.wq_ta.info.addrof("unk_54")),

    "ta_ts": t(work.wc_ta.ts1._addr),
    "ta_tsb": t(workb.wc_ta.ts1._addr),
    "3d_ts": t(work.wc_3d.ts1._addr),
    "3d_tsb": t(workb.wc_3d.ts1._addr),
    "2_ta_ts": t(work2.wc_ta.ts1._addr),
    "2_ta_tsb": t(work2b.wc_ta.ts1._addr),
    "2_3d_ts": t(work2.wc_3d.ts1._addr),
    "2_3d_tsb": t(work2b.wc_3d.ts1._addr),

    "ta_unkts": t(work.wc_ta.unk_ts._addr),
    "ta_unktsb": t(workb.wc_ta.unk_ts._addr),
    "3d_unkts": t(work.wc_3d.unk_ts._addr),
    "3d_unktsb": t(workb.wc_3d.unk_ts._addr),
    "2_ta_unkts": t(work2.wc_ta.unk_ts._addr),
    "2_ta_unktsb": t(work2b.wc_ta.unk_ts._addr),
    "2_3d_unkts": t(work2.wc_3d.unk_ts._addr),
    "2_3d_unktsb": t(work2b.wc_3d.unk_ts._addr),

    "ta_unkts2": t(work.wc_ta.unk_ts2._addr),
    "ta_unkts2b": t(workb.wc_ta.unk_ts2._addr),
    "3d_unkts2": t(work.wc_3d.unk_ts2._addr),
    "3d_unkts2b": t(workb.wc_3d.unk_ts2._addr),
    "2_ta_unkts2": t(work2.wc_ta.unk_ts2._addr),
    "2_ta_unkts2b": t(work2b.wc_ta.unk_ts2._addr),
    "2_3d_unkts2": t(work2.wc_3d.unk_ts2._addr),
    "2_3d_unkts2b": t(work2b.wc_3d.unk_ts2._addr),

    "ta_start": t(work.tsta_start._addr),
    "ta_end": t(work.tsta_end._addr),
    "3d_start": t(work.ts3d_start._addr),
    "3d_end": t(work.ts3d_end._addr),
    "ta2_sart": t(workb.tsta_start._addr),
    "ta2_end": t(workb.tsta_end._addr),
    "3d2_start": t(workb.ts3d_start._addr),
    "3d2_end": t(workb.ts3d_end._addr),
    "2_ta_start": t(work2.tsta_start._addr),
    "2_ta_end": t(work2.tsta_end._addr),
    "2_3d_start": t(work2.ts3d_start._addr),
    "2_3d_end": t(work2.ts3d_end._addr),
    "2_ta2_sart": t(work2b.tsta_start._addr),
    "2_ta2_end": t(work2b.tsta_end._addr),
    "2_3d2_start": t(work2b.ts3d_start._addr),
    "2_3d2_end": t(work2b.ts3d_end._addr),

    "ns_ta_start": t(work.nsta_start._addr),
    "ns_ta_end": t(work.nsta_end._addr),
    "ns_3d_start": t(work.ns3d_start._addr),
    "ns_3d_end": t(work.ns3d_end._addr),
    "ns_ta2_sart": t(workb.nsta_start._addr),
    "ns_ta2_end": t(workb.nsta_end._addr),
    "ns_3d2_start": t(workb.ns3d_start._addr),
    "ns_3d2_end": t(workb.ns3d_end._addr),
    "2_ns_ta_start": t(work2.nsta_start._addr),
    "2_ns_ta_end": t(work2.nsta_end._addr),
    "2_ns_3d_start": t(work2.ns3d_start._addr),
    "2_ns_3d_end": t(work2.ns3d_end._addr),
    "2_ns_ta2_sart": t(work2b.nsta_start._addr),
    "2_ns_ta2_end": t(work2b.nsta_end._addr),
    "2_ns_3d2_start": t(work2b.ns3d_start._addr),
    "2_ns_3d2_end": t(work2b.ns3d_end._addr),

    "ta_stamp1":    t(renderer.stamp_ta1._addr),
    "ta_stamp2":    t(renderer.stamp_ta2._addr),
    "3d_stamp1":    t(renderer.stamp_3d1._addr),
    "3d_stamp2":    t(renderer.stamp_3d2._addr),

    "2_ta_stamp1":    t(renderer2.stamp_ta1._addr),
    "2_ta_stamp2":    t(renderer2.stamp_ta2._addr),
    "2_3d_stamp1":    t(renderer2.stamp_3d1._addr),
    "2_3d_stamp2":    t(renderer2.stamp_3d2._addr),
}

div=4
ticks = 24000000 // div * 3

la = GPIOLogicAnalyzer(u, regs=regs, cpu=analyzer_cpu, div=div)


print("==========================================")
print("## Poll prior to job start")
print("==========================================")

mon.poll()
agx.poll_objects()

print("==========================================")
print("## Run")
print("==========================================")

la.start(ticks, bufsize=0x400000)
renderer.run()

print("==========================================")
print("## After r1 start")
print("==========================================")
#agx.poll_objects()

#time.sleep(0.1)
#mon.poll()
#time.sleep(0.15)
#mon.poll()
renderer2.run()

print("==========================================")
print("## After r2 start")
print("==========================================")
agx.poll_objects()

#mon.poll()
print("==========================================")
print("## Waiting")
print("==========================================")

try:

    #while not work.ev_3d.fired:
        #agx.asc.work()
        ##mon.poll()
        #agx.poll_objects()
        #agx.poll_channels()
        #print("==========================================")
        ##time.sleep(0.1)

    #print("==========================================")
    #print("## Ev1 Fired")
    #print("==========================================")

    while not work2b.ev_3d.fired:
        agx.asc.work()
        #mon.poll()
        agx.poll_objects()
        agx.poll_channels()
        print("==========================================")
        #time.sleep(0.1)

    print("==========================================")
    print("## Ev2 Fired")
    print("==========================================")

    renderer.wait()
    renderer2.wait()

    agx.poll_objects()
    #mon.poll()

finally:
    la.complete()
    la.show()

time.sleep(2)




# SPDX-License-Identifier: MIT
from .base import *
from ...utils import *
from construct import *
from ...sysreg import *

class CrashLogMessage(Register64):
    TYPE = 63, 52
    SIZE = 51, 44
    DVA = 43, 0

CrashHeader = Struct(
    "type" / Const("CLHE", FourCC),
    "ver" / Int32ul,
    "total_size" / Int32ul,
    "flags" / Int32ul,
    Padding(16)
)

CrashCver = Struct(
    "uuid" / Bytes(16),
    "version" / CString("utf8"),
)

CrashCstr = Struct(
    "id" / Int32ul,
    "string" / CString("utf8"),
)

CrashCtim = Struct(
    "time" / Int64ul,
)

CrashCmbx = Struct(
    "hdr" / Array(4, Hex(Int32ul)),
    "type" / Int32ul,
    "unk" / Int32ul,
    "index" / Int32ul,
    "messages" / GreedyRange(Struct(
        "endpoint" / Hex(Int64ul),
        "message" / Hex(Int64ul),
        "timestamp" / Hex(Int32ul),
        Padding(4),
    )),
)

CrashCcst = Struct(
    "task" / Int32ul,
    "unk" / Int32ul,
    "stack" / GreedyRange(Int64ul)
)

CrashCasC = Struct(
    "l2c_err_sts" / Hex(Int64ul),
    "l2c_err_adr" / Hex(Int64ul),
    "l2c_err_inf" / Hex(Int64ul),
    "lsu_err_sts" / Hex(Int64ul),
    "fed_err_sts" / Hex(Int64ul),
    "mmu_err_sts" / Hex(Int64ul)
)

CrashCrg8 = Struct(
    "unk_0" / Int32ul,
    "unk_4" / Int32ul,
    "regs" / Array(31, Hex(Int64ul)),
    "sp" / Int64ul,
    "pc" / Int64ul,
    "psr" / Int64ul,
    "cpacr" / Int64ul,
    "fpsr" / Int64ul,
    "fpcr" / Int64ul,
    "unk" / Array(64, Hex(Int64ul)),
    "far" / Int64ul,
    "unk_X" / Int64ul,
    "esr" / Int64ul,
    "unk_Z" / Int64ul,
)

CrashEntry = Struct(
    "type" / FourCC,
    Padding(4),
    "flags" / Hex(Int32ul),
    "len" / Int32ul,
    "payload" / FixedSized(lambda ctx: ctx.len - 16 if ctx.type != "CLHE" else 16,
                           Switch(this.type, {
        "Cver": CrashCver,
        "Ctim": CrashCtim,
        "Cmbx": CrashCmbx,
        "Cstr": CrashCstr,
        "Crg8": CrashCrg8,
        "Ccst": CrashCcst,
        "CasC": CrashCasC,
    }, default=GreedyBytes)),
)

CrashLog = Struct(
    "header" / CrashHeader,
    "entries" / RepeatUntil(this.type == "CLHE", CrashEntry),
)

class CrashLogParser:
    def __init__(self, data=None, asc=None):
        self.asc = asc
        if data is not None:
            self.parse(data)

    def parse(self, data):
        self.data = CrashLog.parse(data)
        pass

    def default(self, entry):
        print(f"# {entry.type} flags={entry.flags:#x}")
        chexdump(entry.payload)
        print()

    def Ccst(self, entry):
        print(f"Call stack (task {entry.payload.task}:")
        for i in entry.payload.stack:
            if not i:
                break
            print(f"  - {i:#x}")
        print()

    def CasC(self, entry):
        print(f"Async error info:")
        print(entry.payload)
        print()

    def Cver(self, entry):
        print(f"RTKit Version: {entry.payload.version}")
        print()

    def Crg8(self, entry):
        print(f"Exception info:")

        ctx = entry.payload

        addr = self.asc.addr

        spsr = SPSR(ctx.psr)
        esr = ESR(ctx.esr)
        elr = ctx.pc
        far_phys = self.asc.iotranslate(ctx.far, 1)[0][0]
        elr_phys = self.asc.iotranslate(ctx.pc, 1)[0][0]
        sp_phys = self.asc.iotranslate(ctx.sp, 1)[0][0]

        print(f"  == Exception taken from {spsr.M.name} ==")
        el = spsr.M >> 2
        print(f"  SPSR   = {spsr}")
        print(f"  ELR    = {addr(elr)}" + (f" (0x{elr_phys:x})" if elr_phys else ""))
        print(f"  ESR    = {esr}")
        print(f"  FAR    = {addr(ctx.far)}" + (f" (0x{far_phys:x})" if far_phys else ""))
        print(f"  SP     = {ctx.sp:#x}" + (f" (0x{sp_phys:x})" if sp_phys else ""))

        for i in range(0, 31, 4):
            j = min(30, i + 3)
            print(f"  {f'x{i}-x{j}':>7} = {' '.join(f'{r:016x}' for r in ctx.regs[i:j + 1])}")

        if elr_phys:
            v = self.asc.p.read32(elr_phys)

            print()
            if v == 0xabad1dea:
                print("  == Faulting code is not available ==")
            else:
                print("  == Faulting code ==")
                dist = 16
                self.asc.u.disassemble_at(elr_phys - dist * 4, (dist * 2 + 1) * 4, elr_phys)

        print()

    def Cstr(self, entry):
        print(f"Message {entry.payload.id}: {entry.payload.string}")
        print()

    def Ctim(self, entry):
        print(f"Crash time: {entry.payload.time:#x}")
        print()

    def Cmbx(self, entry):
        print(f"Mailbox log (type {entry.payload.type}, index {entry.payload.index}):")
        for i, msg in enumerate(entry.payload.messages):
            print(f" #{i:3d} @{msg.timestamp:#10x} ep={msg.endpoint:#4x} {msg.message:#18x}")
        print()

    def CLHE(self, entry):
        pass

    def dump(self):
        print("### Crash dump:")
        print()
        for entry in self.data.entries:
            getattr(self, entry.type, self.default)(entry)

class ASCCrashLogEndpoint(ASCBaseEndpoint):
    SHORT = "crash"
    BASE_MESSAGE = CrashLogMessage

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.iobuffer = None
        self.iobuffer_dva = None
        self.started = False

    @msg_handler(0x1)
    def Handle(self, msg):
        if self.started:
            return self.handle_crashed(msg)
        else:
            return self.handle_getbuf(msg)

    def handle_getbuf(self, msg):

        if msg.DVA:
            size = 0x1000 * msg.SIZE
            self.iobuffer_dva = msg.DVA
            self.log(f"buf prealloc at dva {self.iobuffer_dva:#x}")
        else:
            size = align(0x1000 * msg.SIZE, 0x4000)
            self.iobuffer, self.iobuffer_dva = self.asc.ioalloc(size)
            self.log(f"buf {self.iobuffer:#x} / {self.iobuffer_dva:#x}")
            self.send(CrashLogMessage(TYPE=1, SIZE=size // 0x1000, DVA=self.iobuffer_dva))

        self.started = True
        return True

    def crash_soft(self):
        self.send(0x40)

    def crash_hard(self):
        self.send(0x22)

    def handle_crashed(self, msg):
        size = 0x1000 * msg.SIZE

        self.log(f"Crashed!")
        crashdata = self.asc.ioread(msg.DVA, size)
        open("crash.bin", "wb").write(crashdata)
        clog = CrashLogParser(crashdata, self.asc)
        clog.dump()
        raise Exception("ASC crashed!")

        return True

if __name__ == "__main__":
    import sys
    crashdata = open(sys.argv[1], "rb").read()
    clog = CrashLogParser(crashdata)
    clog.dump()



# SPDX-License-Identifier: MIT
from construct import *
from enum import IntEnum

from ..utils import *

__all__ = [
    "MMIOTraceFlags", "EvtMMIOTrace", "EvtIRQTrace", "HV_EVENT",
    "VMProxyHookData", "TraceMode",
]

class MMIOTraceFlags(Register32):
    ATTR = 31, 24
    CPU = 23, 16
    SH = 15, 14
    WIDTH = 4, 0
    WRITE = 5
    MULTI = 6

EvtMMIOTrace = Struct(
    "flags" / RegAdapter(MMIOTraceFlags),
    "reserved" / Int32ul,
    "pc" / Hex(Int64ul),
    "addr" / Hex(Int64ul),
    "data" / Hex(Int64ul),
)

EvtIRQTrace = Struct(
    "flags" / Int32ul,
    "type" / Hex(Int16ul),
    "num" / Int16ul,
)

class HV_EVENT(IntEnum):
    HOOK_VM = 1
    VTIMER = 2
    USER_INTERRUPT = 3
    WDT_BARK = 4
    CPU_SWITCH = 5
    VIRTIO = 6
    PANIC = 7

VMProxyHookData = Struct(
    "flags" / RegAdapter(MMIOTraceFlags),
    "id" / Int32ul,
    "addr" / Hex(Int64ul),
    "data" / Array(8, Hex(Int64ul)),
)

class TraceMode(IntEnum):
    '''
Different types of Tracing '''

    OFF = 0
    BYPASS = 1
    ASYNC = 2
    UNBUF = 3
    WSYNC = 4
    SYNC = 5
    HOOK = 6
    RESERVED = 7



# SPDX-License-Identifier: MIT
import struct
from dataclasses import dataclass
from enum import IntEnum

from ..asc.base import *
from ...utils import *

## DCP main endpoint

class DCPMessage(Register64):
    TYPE        = 3, 0

class DCPEp_SetShmem(DCPMessage):
    DVA         = 63, 16
    FLAG        = 7, 4
    TYPE        = 3, 0, Constant(0)

class DCPEp_InitComplete(DCPMessage):
    TYPE        = 3, 0, Constant(1)

class CallContext(IntEnum):
    CB          = 0
    CMD         = 2
    ASYNC       = 3
    OOBCB       = 4
    OOBCMD      = 6
    OOBASYNC    = 7

class DCPEp_Msg(DCPMessage):
    LEN         = 63, 32
    OFF         = 31, 16
    CTX         = 11, 8, CallContext
    ACK         = 6
    TYPE        = 3, 0, Constant(2)

@dataclass
class DCPCallState:
    tag: str
    off: int
    in_len: int
    in_data: bytes
    out_addr: int
    out_len: int
    complete: bool = False

class DCPCallChannel(Reloadable):
    def __init__(self, dcpep, name, buf, bufsize):
        self.dcp = dcpep
        self.name = name
        self.buf = buf
        self.bufsize = bufsize
        self.off = 0
        self.pending = []

    def ack(self):
        if not self.pending:
            raise Exception("ACK with no calls pending")

        self.pending[-1].complete = True

    def call(self, ctx, tag, inbuf, out_len):
        in_len = len(inbuf)
        data = tag.encode("ascii")[::-1] + struct.pack("<II", in_len, out_len) + inbuf
        data_size = len(data) + out_len
        assert (self.off + data_size) <= self.bufsize

        self.dcp.asc.iface.writemem(self.dcp.shmem + self.buf + self.off, data)

        state = DCPCallState(off=self.off, tag=tag, in_len=in_len, in_data=data, out_len=out_len,
                             out_addr=self.buf + self.off + 12 + in_len)

        self.off += align_up(data_size, 0x40)
        self.pending.append(state)

        print(f"len={data_size:#x} {in_len}")
        self.dcp.send(DCPEp_Msg(LEN=data_size, OFF=state.off, CTX=ctx, ACK=0))

        while not state.complete:
            self.dcp.asc.work()

        print(f"off={state.out_addr:#x} len={out_len}")
        out_data = self.dcp.asc.iface.readmem(self.dcp.shmem + state.out_addr, out_len)

        assert self.pending.pop() is state
        self.off = state.off

        return out_data

class DCPCallbackChannel(Reloadable):
    def __init__(self, dcpep, name, buf, bufsize):
        self.dcp = dcpep
        self.name = name
        self.buf = buf
        self.bufsize = bufsize
        self.pending = []

    def cb(self, msg):
        data = self.dcp.asc.iface.readmem(self.dcp.shmem + self.buf + msg.OFF, msg.LEN)
        tag = data[:4][::-1].decode("ascii")
        in_len, out_len = struct.unpack("<II", data[4:12])
        in_data = data[12:12 + in_len]

        state = DCPCallState(off=msg.OFF, tag=tag, in_len=in_len, out_len=out_len,
                             in_data=in_data, out_addr=self.buf + msg.OFF + 12 + in_len)

        self.pending.append(state)

        out_data = self.dcp.mgr.handle_cb(state)
        self.dcp.asc.iface.writemem(self.dcp.shmem + state.out_addr, out_data)
        self.dcp.send(DCPEp_Msg(CTX=msg.CTX, ACK=1))

        assert self.pending.pop() is state


class DCPEndpoint(ASCBaseEndpoint):
    BASE_MESSAGE = DCPMessage
    SHORT = "dcpep"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.shmem = self.shmem_dva = None
        self.init_complete = False
        self.mgr = None

        self.ch_cb = DCPCallbackChannel(self, "CB", 0x60000, 0x8000)
        self.ch_cmd = DCPCallChannel(self, "CMD", 0, 0x8000)
        self.ch_async = DCPCallbackChannel(self, "ASYNC", 0x40000, 0x8000)
        self.ch_oobasync = DCPCallbackChannel(self, "OOBASYNC", 0x48000, 0x8000)
        self.ch_oobcb = DCPCallbackChannel(self, "OOBCB", 0x68000, 0x8000)
        self.ch_oobcmd = DCPCallChannel(self, "OOBCMD", 0x8000, 0x8000)

    @msg_handler(2, DCPEp_Msg)
    def Rx(self, msg):
        if msg.ACK:
            if msg.CTX in (CallContext.CMD, CallContext.CB):
                self.ch_cmd.ack()
            elif msg.CTX in (CallContext.OOBCMD, CallContext.OOBCB):
                self.ch_oobcmd.ack()
            else:
                raise Exception(f"Unknown RX ack channel {msg.CTX}")
        else:
            if msg.CTX == CallContext.CB:
                self.ch_cb.cb(msg)
            elif msg.CTX == CallContext.OOBCMD:
                self.ch_oobcb.cb(msg)
            elif msg.CTX == CallContext.ASYNC:
                self.ch_async.cb(msg)
            elif msg.CTX == CallContext.OOBASYNC:
                self.ch_oobasync.cb(msg)
            else:
                raise Exception(f"Unknown RX callback channel {msg.CTX}")
        return True

    @msg_handler(1, DCPEp_InitComplete)
    def InitComplete(self, msg):
        self.log("init complete")
        self.init_complete = True
        return True

    def initialize(self):
        self.shmem, self.shmem_dva = self.asc.ioalloc(0x100000)
        self.asc.p.memset32(self.shmem, 0, 0x100000)
        self.send(DCPEp_SetShmem(DVA=self.shmem_dva))
        while not self.init_complete:
            self.asc.work()




# SPDX-License-Identifier: MIT

import struct

from ..common import *
from ...utils import *
from ..asc.base import *


class AFKEPMessage(Register64):
    TYPE = 63, 48

class AFKEP_GetBuf(AFKEPMessage):
    TYPE = 63, 48, Constant(0x89)
    SIZE = 31, 16
    TAG = 15, 0

class AFKEP_GetBuf_Ack(AFKEPMessage):
    TYPE = 63, 48, Constant(0xa1)
    DVA = 47, 0

class AFKEP_InitRB(AFKEPMessage):
    OFFSET = 47, 32
    SIZE = 31, 16
    TAG = 15, 0

class AFKEP_Send(AFKEPMessage):
    TYPE = 63, 48, Constant(0xa2)
    WPTR = 31, 0

class AFKEP_Recv(AFKEPMessage):
    TYPE = 63, 48, Constant(0x85)
    WPTR = 31, 0

class AFKEP_Init(AFKEPMessage):
    TYPE = 63, 48, Constant(0x80)

class AFKEP_Init_Ack(AFKEPMessage):
    TYPE = 63, 48, Constant(0xa0)

class AFKEP_Start(AFKEPMessage):
    TYPE = 63, 48, Constant(0xa3)

class AFKEP_Start_Ack(AFKEPMessage):
    TYPE = 63, 48, Constant(0x86)

class AFKEP_Shutdown(AFKEPMessage):
    TYPE = 63, 48, Constant(0xc0)

class AFKEP_Shutdown_Ack(AFKEPMessage):
    TYPE = 63, 48, Constant(0xc1)


class AFKError(Exception):
    pass

"""
The first three blocks of the ringbuffer is reserved for exchanging size,
rptr, wptr:

           bufsize      unk
00000000  00007e80 00070006 00000000 00000000 00000000 00000000 00000000 00000000
00000020  00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000
00000040  *   rptr
00000080  00000600 00000000 00000000 00000000 00000000 00000000 00000000 00000000
000000a0  00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000
000000c0  *   wptr
00000100  00000680 00000000 00000000 00000000 00000000 00000000 00000000 00000000
00000120  00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000
00000140  *

Note how each block is spread out by some block_size multiple of 0x40
(step).  Here, the block_size is 0x80. The 0th block holds the bufsize,
the 1st block holds the rptr, and the 2nd block holds the wptr. The
actual contents of the ringbuffer starts after the first three blocks,
which will be called the "header". Since we're *always* given the total
block size at offset +0x0 or +block_size*0, we can calculate the block
size by dividing by 3.
"""

class AFKRingBuf(Reloadable):
    BLOCK_STEP = 0x40
    BLOCK_COUNT = 3

    def __init__(self, ep, base, size):
        self.ep = ep
        self.base = base

        bs, unk = struct.unpack("<II", self.read_buf(0, 8))
        # calculate block_size
        # bs + self.BLOCK_COUNT * block_size) == size
        assert((size - bs) % self.BLOCK_COUNT == 0)
        block_size = (size - bs) // self.BLOCK_COUNT
        assert(block_size % self.BLOCK_STEP == 0)
        self.block_size = block_size
        self.bufsize = bs
        self.rptr = 0
        self.wptr = 0

    def read_buf(self, off, size):
        return self.ep.iface.readmem(self.base + off, size)

    def write_buf(self, off, data):
        return self.ep.iface.writemem(self.base + off, data)
    
    def get_rptr(self):
        return struct.unpack("<I", self.read_buf(self.block_size * 1, 4))[0]
        #return self.ep.asc.p.read32(self.base + self.BLOCK_STEP)

    def get_wptr(self):
        return struct.unpack("<I", self.read_buf(self.block_size * 2, 4))[0]
        #return self.ep.asc.p.read32(self.base + 2 * self.BLOCK_STEP)

    def update_rptr(self, rptr):
        self.write_buf(self.block_size * 1, struct.pack("<I", rptr))
        self.ep.asc.p.write32(self.base + self.BLOCK_STEP, rptr)

    def update_wptr(self, wptr):
        self.write_buf(self.block_size * 2, struct.pack("<I", wptr))
        self.ep.asc.p.write32(self.base + 2 * self.BLOCK_STEP, wptr)

    def read(self):
        self.wptr = self.get_wptr()

        base = self.block_size * 3  # after header (size, rptr, wptr)
        while self.wptr != self.rptr:
            hdr = self.read_buf(base + self.rptr, 16)
            self.rptr += 16
            magic, size = struct.unpack("<4sI", hdr[:8])
            assert magic in [b"IOP ", b"AOP "]
            if size > (self.bufsize - self.rptr):
                hdr = self.read_buf(base, 16)
                self.rptr = 16
                magic, size = struct.unpack("<4sI", hdr[:8])
                assert magic in [b"IOP ", b"AOP "]

            payload = self.read_buf(base + self.rptr, size)
            self.rptr = (align_up(self.rptr + size, self.block_size)) % self.bufsize
            self.update_rptr(self.rptr)
            yield hdr[8:] + payload
            self.wptr = self.get_wptr()

        self.update_rptr(self.rptr)

    def write(self, data):
        base = self.block_size * 3  # after header (size, rptr, wptr)
        hdr2, data = data[:8], data[8:]
        self.rptr = self.get_rptr()
        
        if self.wptr < self.rptr and self.wptr + 0x10 >= self.rptr:
            raise AFKError("Ring buffer is full")

        hdr = struct.pack("<4sI", b"IOP ", len(data)) + hdr2
        self.write_buf(base + self.wptr, hdr)

        if len(data) > (self.bufsize - self.wptr - 16):
            if self.rptr < 0x10:
                raise AFKError("Ring buffer is full")
            self.write_buf(base, hdr)
            self.wptr = 0

        if self.wptr < self.rptr and self.wptr + 0x10 + len(data) >= self.rptr:
            raise AFKError("Ring buffer is full")

        self.write_buf(base + self.wptr + 0x10, data)
        self.wptr = align_up(self.wptr + 0x10 + len(data), self.block_size) % self.bufsize

        self.update_wptr(self.wptr)
        return self.wptr

class AFKRingBufEndpoint(ASCBaseEndpoint):
    BASE_MESSAGE = AFKEPMessage
    SHORT = "afkep"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.txq = None
        self.rxq = None
        self.iface = self.asc.iface
        self.alive = False
        self.started = False
        self.iobuffer = None
        self.verbose = 2
        self.msgid = 0

    def start(self):
        self.send(AFKEP_Init())

    def stop(self):
        self.log("Shutting down")
        self.send(AFKEP_Shutdown())
        while self.alive:
            self.asc.work()

    @msg_handler(0xa0, AFKEP_Init_Ack)
    def Init_Ack(self, msg):
        self.alive = True
        return True

    @msg_handler(0xc1, AFKEP_Shutdown_Ack)
    def Shutdown_Ack(self, msg):
        self.alive = False
        self.log("Shutdown ACKed")
        return True

    @msg_handler(0x80, AFKEP_Init)
    def Hello(self, msg):
        self.rxbuf, self.rxbuf_dva = self.asc.ioalloc(0x10000)
        self.txbuf, self.txbuf_dva = self.asc.ioalloc(0x10000)
        self.send(AFKEP_Init_Ack())
        return True

    @msg_handler(0x89, AFKEP_GetBuf)
    def GetBuf(self, msg):
        size = msg.SIZE * AFKRingBuf.BLOCK_STEP

        if self.iobuffer:
            print("WARNING: trying to reset iobuffer!")

        self.iobuffer, self.iobuffer_dva = self.asc.ioalloc(size)
        self.asc.p.write32(self.iobuffer, 0xdeadbeef)
        self.send(AFKEP_GetBuf_Ack(DVA=self.iobuffer_dva))
        self.log(f"Buffer: phys={self.iobuffer:#x} dva={self.iobuffer_dva:#x} size={size:#x}")
        return True

    @msg_handler(0x8a, AFKEP_InitRB)
    def InitTX(self, msg):
        self.txq = self.init_rb(msg)
        if self.rxq and self.txq:
            self.start_queues()
        return True

    @msg_handler(0x8b, AFKEP_InitRB)
    def InitRX(self, msg):
        self.rxq = self.init_rb(msg)
        if self.rxq and self.txq:
            self.start_queues()
        return True

    @msg_handler(0x8c, AFKEP_InitRB)
    def InitUnk(self, msg):
        return True  # no op

    def init_rb(self, msg):
        off = msg.OFFSET * AFKRingBuf.BLOCK_STEP
        size = msg.SIZE * AFKRingBuf.BLOCK_STEP
        return AFKRingBuf(self, self.iobuffer + off, size)

    def start_queues(self):
        self.send(AFKEP_Start())

    @msg_handler(0x86, AFKEP_Start_Ack)
    def Start_Ack(self, msg):
        self.started = True
        return True

    @msg_handler(0x85, AFKEP_Recv)
    def Recv(self, msg):
        for data in self.rxq.read():
            if self.verbose >= 3:
                self.log(f"<RX rptr={self.rxq.rptr:#x}")
                chexdump(data)
            self.handle_ipc(data)
        return True

    def handle_ipc(self, data):
        pass

    def send_ipc(self, data):
        wptr = self.txq.write(data)
        self.send(AFKEP_Send(WPTR = wptr))



# SPDX-License-Identifier: MIT
from m1n1.trace import Tracer
from m1n1.trace.dart import DARTTracer
from m1n1.utils import *
from m1n1.proxyutils import RegMonitor
hv.p.hv_set_time_stealing(0, 1)

# Usage
#
# 2023/12/25: Only tested on J293AP (AVD revision "Viola"/V3).
#             Should work on all baseline M1s.
#
# 1. Start tracer under the hypervisor
#
# 2. Send over the bitstream(s) to the target machine
#    Supported formats: .264, .265, .ivf (unless you want to add a demuxer to avid/codecs)
#
# 3. The tracer is purposely not activated at boot. As of 13.5, it takes ~2 HEVC runs
#    to get to the login screen - it's probably decoding the login screen. By "activate",
#    I mean the tracer variable "outdir" is NULLed s.t. the tracer will not save the
#    traced data and merely log the IPC transactions.
#
#       [cpu0] [AVDTracer@/arm-io/avd] sent fw command at 0x108eb30
#       00000000  00000801 00030007 0050c000 000002a8 00000003 01091a70 01091a78 00000001
#       00000020  0050caa4 01091af0 01091bc0 01091c50 0050c2a4 0050c210 0050c28c 00000000
#       00000040  00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000
#       [cpu0] [0xfffffe00149d9984] MMIO: W.4   0x269098054 (avd[0], offset 0x1098054) = 0x108eb30
#       [cpu7] [0xfffffe00149d99b0] MMIO: R.4   0x269098048 (avd[0], offset 0x1098048) = 0x9
#
# 4. To save the trace contents, break into the hypervisor console and set
#
#       >>> tracer.outdir = "matrix_1080X512"
#
#    The output data dir will become 'data/[inferred codec name]/$outdir'.
#    The data dir will be created if it does not exist. The directory structure is
#    meant to look like this:
#
#    proxyclient/
#       data/
#          h264/*
#          h265/*
#          vp9/*
#
# 5. After the directory name is configured, trigger avd from the target machine:
#
#       ffmpeg -hwaccel videotoolbox -i matrix_1080X512.ivf
#
#    Or you can access VT directly yourself with some Obj-C (but why would you want to do that..)
#    Though I may need to do that to test some sekrit undocumented features.
#    The input bitstream is hopefully the matching one as the directory name just set.
#
# 6. If all goes well (i.e. the bitstream is decoding on AVD), the tracer will save:
#
#       >> ~/m1n1/proxyclient $ ls data/*
#       data/vp9/matrix_1080X512:
#       frame.2023-12-17T21:17:47.519048.00004000.bin  probs.2023-12-17T21:17:47.519048.00004000.00004000.bin
#       frame.2023-12-17T21:17:47.578537.000bc000.bin  probs.2023-12-17T21:17:47.578537.000bc000.0000c000.bin
#       frame.2023-12-17T21:17:47.633768.00174000.bin  probs.2023-12-17T21:17:47.633768.00174000.00014000.bin
#       frame.2023-12-17T21:17:47.688067.0022c000.bin  probs.2023-12-17T21:17:47.688067.0022c000.0001c000.bin
#
#   The "frame" is the macOS source frame_params struct, and this directory is the
#   one intended to be supplied to all the avid tools (e.g. emulator, differs).
#   For VP9 (and presumably AV1) the tracer will additionally save the "probs" blob.
#   You can also bypass FairPlay encryption and save the coded bitstream, but that's
#   an exercise left for the reader.
#
# 7. Copy the data directory over to `avd/data/*` & have fun :D
#
#    python3 avd_emu.py -i frame.2023-12-17T21:17:47.519048.00004000.bin  # emulate single fp
#    python3 avd_emu.py -d vp9/matrix_1080X512 -a  # emulate all using trace dir name
#    python3 tools/test.py -m vp9 -d vp9/matrix_1080X512 -e -a  # test against emulated output
#
#    Optionally save the firmware to run on the emulator via
#       >>> tracer.save_firmware("data/fw.bin")

import datetime
import os
import struct
import time

class AVDTracer(Tracer):
    DEFAULT_MODE = TraceMode.SYNC

    def __init__(self, hv, dev_path, dart_tracer, verbose=False):
        super().__init__(hv, verbose=verbose, ident=type(self).__name__ + "@" + dev_path)
        self.dev = hv.adt[dev_path]
        self.dart_tracer = dart_tracer
        self.base = self.dev.get_reg(0)[0] # 0x268000000
        self.p = hv.p
        self.u = hv.u
        self.dart = dart_tracer.dart

        mon = RegMonitor(hv.u)
        AVD_REGS = [
            #(0x1000000, 0x4000, "unk0"),
            #(0x1010000, 0x4000, "dart"),
            #(0x1002000, 0x1000, "unk2"),
            (0x1070000, 0x4000, "piodma"),
            (0x1088000, 0x4000, "sram"),
            (0x108c000, 0xc000, "cmd"),
            #(0x1098000, 0x4000, "mbox"),
            #(0x10a3000, 0x1000, "unka"),
            (0x1100000, 0xc000, "config"),
            (0x110c000, 0x4000, "dma"),
            #(0x1400000, 0x4000, "wrap"),
        ]
        #for (offset, size, name) in AVD_REGS: mon.add(self.base + offset, size, name=name)
        self.mon = mon

        iomon = RegMonitor(hv.u, ascii=True)
        iomon1 = RegMonitor(hv.u, ascii=True)
        def readmem_iova(addr, size, readfn=None):
            try:
                return dart_tracer.dart.ioread(0, addr, size)
            except Exception as e:
                print(e)
                return None
        iomon.readmem = readmem_iova
        def readmem_iova(addr, size, readfn=None):
            try:
                return dart_tracer.dart.ioread(1, addr, size)
            except Exception as e:
                print(e)
                return None
        iomon1.readmem = readmem_iova
        self.iomon = iomon
        self.iomon1 = iomon1
        self.state_active = False
        self.outdir = ""

    def avd_r32(self, off): return self.p.read32(self.base + off)
    def avd_w32(self, off, x): return self.p.write32(self.base + off, x)
    def avd_r64(self, off): return self.p.read64(self.base + off)
    def avd_w64(self, off, x): return self.p.write64(self.base + off, x)

    def start(self):
        self.hv.trace_range(irange(self.dev.get_reg(0)[0], self.dev.get_reg(0)[1]), mode=TraceMode.SYNC)
        self.hv.trace_range(irange(self.base + 0x1080000, 0x18000), False)
        self.hv.add_tracer(irange(self.base + 0x1098054, 4), "avd-mbox-54", TraceMode.SYNC, self.evt_rw_hook, self.w_AVD_MBOX_0054)
        self.hv.add_tracer(irange(self.base + 0x1098064, 4), "avd-mbox-64", TraceMode.SYNC, self.r_AVD_MBOX_0064, self.evt_rw_hook)

    def poll(self):
        self.mon.poll()
        self.iomon.poll()
        self.iomon1.poll()

    def evt_rw_hook(self, x):
        self.poll()

    def w_AVD_MBOX_0054(self, x):
        if ((x.data >= 0x1080000) and (x.data <= 0x10a0000)):
            self.log("Sent fw command at 0x%x" % (x.data))
            self.poll()
            cmd = self.read_regs(self.base + x.data, 0x60)
            chexdump32(cmd)

            opcode = struct.unpack("<I", cmd[:4])[0] & 0xf
            if (opcode == 0):
                self.log("Command start")
                self.state_active = True
                self.access_idx = 0

            elif (opcode == 1):
                frame_params_iova = self.p.read32(self.base + x.data + 0x8)
                if (self.outdir) and (frame_params_iova != 0x0):
                    t = datetime.datetime.now().isoformat()
                    frame_params = self.dart.ioread(1, frame_params_iova, 0xb0000)

                    word = self.p.read32(self.base + x.data)
                    if   (word & 0x000) == 0x000: # h265
                        name = "h265"
                    elif (word & 0x400) == 0x400: # h264
                        name = "h264"
                    elif (word & 0x800) == 0x800: # vp9
                        name = "vp9"
                    else:
                        name = "unk"
                    outdir = os.path.join("data", name, self.outdir)
                    os.makedirs(outdir, exist_ok=True)
                    open(os.path.join(outdir, f'frame.{t}.{frame_params_iova:08x}.bin'), "wb").write(frame_params)

                    if (word & 0x800) == 0x800: # save probs for vp9
                        iova = [0x4000, 0xc000, 0x14000, 0x1c000][self.access_idx % 4]
                        open(os.path.join(outdir, f'probs.{t}.{frame_params_iova:08x}.{iova:08x}.bin'), "wb").write(self.dart.ioread(0, iova, 0x4000))
                self.access_idx += 1

            elif (opcode == 2):
                self.log("Command end")
                self.state_active = False
                self.access_idx = 0

    def r_AVD_MBOX_0064(self, x):
        if ((x.data >= 0x1080000) and (x.data <= 0x10a0000)):
            self.log("Received fw command at 0x%x" % (x.data))
            cmd = self.read_regs(self.base + x.data, 0x60)
            chexdump32(cmd)
            self.poll()

    def read_regs(self, addr, size):
        scratch = self.u.malloc(size)
        p.memcpy32(scratch, addr, size)
        return self.p.iface.readmem(scratch, size)

    def read_iova(self, start, end, stream=0):
            data = b''
            for i in range((end - start) // 0x4000):
                try:
                    d = self.dart_tracer.dart.ioread(stream, start + (i * 0x4000), 0x4000)
                except:
                    d = b'\0' * 0x4000
                data += d
            return data

    def save_firmware(self, path="fw.bin"):
        firmware = self.read_regs(self.base + 0x1080000, 0x10000)
        open(path, "wb").write(firmware)

p.pmgr_adt_clocks_enable('/arm-io/dart-avd')
p.pmgr_adt_clocks_enable('/arm-io/avd')
dart_tracer = DARTTracer(hv, "/arm-io/dart-avd", verbose=0)
dart_tracer.start()
dart = dart_tracer.dart
tracer = AVDTracer(hv, '/arm-io/avd', dart_tracer, verbose=3)
tracer.start()



# SPDX-License-Identifier: MIT

from construct import *
from ..fw.agx.channels import *
from ..fw.agx.cmdqueue import *

class GPUChannel:
    STATE_FIELDS = ChannelStateFields

    def __init__(self, agx, name, channel_id, state_addr, ring_addr, ring_size):
        self.agx = agx
        self.u = agx.u
        self.name = name
        self.channel_id = channel_id
        self.iface = agx.u.iface
        self.state_addr = state_addr
        self.ring_addr = ring_addr
        self.ring_size = ring_size
        self.state = self.STATE_FIELDS(self.u, self.state_addr)
        self.state.READ_PTR.val = 0
        self.state.WRITE_PTR.val = 0

    @classmethod
    def item_size(cls):
        return cls.MSG_CLASS.sizeof()

    def log(self, msg):
        self.agx.log(f"[{self.name}] {msg}")

class GPUTXChannel(GPUChannel):
    def doorbell(self):
        self.agx.asc.db.doorbell(self.channel_id)

    def send_message(self, msg):
        wptr = self.state.WRITE_PTR.val
        self.iface.writemem(self.ring_addr + self.item_size() * wptr,
                            msg.build())
        self.state.WRITE_PTR.val = (wptr + 1) % self.ring_size
        self.doorbell()

class GPURXChannel(GPUChannel):
    def poll(self):
        wptr = self.state.WRITE_PTR.val
        rptr = self.state.READ_PTR.val

        if wptr >= self.ring_size:
            raise Exception(f"wptr = {wptr:#x} > {self.ring_size:#x}")

        while rptr != wptr:
            msg = self.iface.readmem(self.ring_addr + self.item_size() * rptr,
                                     self.item_size())
            self.handle_message(self.MSG_CLASS.parse(msg))
            rptr = (rptr + 1) % self.ring_size
        self.state.READ_PTR.val = rptr

    def handle_message(self, msg):
        self.log(f"Message: {msg}")

class GPUCmdQueueChannel(GPUTXChannel):
    MSG_CLASS = RunCmdQueueMsg

    def run(self, queue, event):
        msg = RunCmdQueueMsg()
        msg.queue_type = queue.TYPE
        msg.cmdqueue = queue.info
        msg.cmdqueue_addr = queue.info._addr
        msg.head = queue.wptr
        msg.event_number = event
        msg.new_queue = 1 if queue.first_time else 0
        msg.timestamp = 0
        queue.first_time = False
        #print(msg)
        self.send_message(msg)

class GPUDeviceControlChannel(GPUTXChannel):
    MSG_CLASS = DeviceControlMsg

    def send_init(self):
        self.send_message(DC_Init())

    def dc_09(self, a, ptr, b):
        # Writes to InitData.RegionB
        msg = DC_09()
        msg.unk_4 = a
        msg.unkptr_c = ptr
        msg.unk_14 = b
        self.send_message(msg)

    def send_foo(self, t, d=None):
        msg = DC_Any()
        msg.msg_type = t
        if d is not None:
            msg.data = d
        self.send_message(msg)

    def update_idle_ts(self):
        self.send_message(DC_UpdateIdleTS())

    def destroy_context(self, ctx):
        msg = DC_DestroyContext()
        msg.unk_4 = 0
        msg.unk_8 = 2
        msg.unk_c = 0
        msg.unk_10 = 0
        msg.unk_14 = 0xffff
        msg.unk_18 = 0
        msg.context_addr = ctx.gpu_context._addr
        print(msg)
        self.send_message(msg)

    # Maybe related to stamps?
    def write32(self, addr, val):
        msg = DC_Write32()
        msg.addr = addr
        msg.data = val
        msg.unk_10 = 0
        msg.unk_14 = 0
        msg.unk_18 = 0
        msg.unk_1c = 0
        print(msg)
        self.send_message(msg)

    def dc_1e(self, a, b):
        msg = DC_1e()
        msg.unk_4 = a
        msg.unk_c = b
        print(msg)
        self.send_message(msg)

class GPUFWCtlChannel(GPUTXChannel):
    STATE_FIELDS = FWControlStateFields
    MSG_CLASS = FWCtlMsg

    def doorbell(self):
        self.agx.asc.db.fwctl_doorbell()

    def send_inval(self, ctx, addr=0):
        msg = FWCtlMsg()
        msg.addr = addr
        msg.unk_8 = 0
        msg.context_id = ctx
        msg.unk_10 = 1
        msg.unk_12 = 2
        print(msg)
        self.send_message(msg)

class GPUEventChannel(GPURXChannel):
    MSG_CLASS = EventMsg

    def handle_message(self, msg):
        if isinstance(msg, FlagMsg):
            self.agx.event_mgr.fired(msg.firing)
        elif isinstance(msg, FaultMsg):
            self.agx.faulted(msg)
        elif isinstance(msg, TimeoutMsg):
            self.agx.timeout(msg)
        else:
            self.log(f"Unknown event: {msg}")

class GPULogChannel(GPURXChannel):
    MSG_CLASS = FWLogMsg

    def handle_message(self, msg):
        ts = msg.timestamp / 24000000
        self.log(f"[{msg.seq_no:<4d}{ts:14.7f}] {msg.msg}")

class GPUKTraceChannel(GPURXChannel):
    MSG_CLASS = KTraceMsg

    def handle_message(self, msg):
        self.log(f"{msg}")

class GPUStatsChannel(GPURXChannel):
    MSG_CLASS = HexDump(Bytes(0x60))

    def handle_message(self, msg):
        if self.agx.show_stats:
            self.log(f"stat {msg}")
